from __future__ import annotations

from typing import Protocol

from ynab_unlinked.models import Transaction, TransactionWithYnabData

from .migrations import Versioned
from .models.shared import EntityConfig


class Config(Versioned, Protocol):
    @staticmethod
    def exists() -> bool: ...
    def save(self): ...
    @staticmethod
    def load() -> Config: ...
    def update_and_save(self, last_transaction: Transaction, entity_name: str): ...
    def add_payee_rules(self, transactions: list[TransactionWithYnabData]): ...
    def payee_from_fules(self, payee: str) -> str | None: ...
    def entity(self, name: str) -> EntityConfig | None: ...
    def set_entity_account(self, name: str, account_id: str): ...



from pathlib import Path
from typing import Annotated

import typer

from .bbva import BBVA


def command(
    context: typer.Context,
    input_file: Annotated[
        Path,
        typer.Argument(exists=True, file_okay=True, dir_okay=False, readable=True),
    ],
):
    """
    Import transactions from BBVA using a Credit Card PDF report.
    """
    from ynab_unlinked.context_object import YnabUnlinkedContext
    from ynab_unlinked.process import process_transactions

    ctx: YnabUnlinkedContext = context.obj

    process_transactions(
        BBVA(),
        input_file,
        ctx,
    )



from enum import StrEnum

from ._protocol import Entity


class InputType(StrEnum):
    TXT = "txt"
    CSV = "csv"
    HTML = "html"
    XLS = "xls"
    XLSX = "xlsx"
    PDF = "pdf"


__all__ = ["Entity", "InputType"]



from __future__ import annotations

from typing import TYPE_CHECKING, assert_never, cast

from ynab_unlinked.entities import Entity, InputType

if TYPE_CHECKING:
    from pathlib import Path

    from ynab_unlinked.context_object import YnabUnlinkedContext
    from ynab_unlinked.models import Transaction


XLSX_ROW_TO_READ = ["", "Fecha", "Tarjeta", "Concepto", "Importe", "Divisa", ""]
VALID_TYPES = [InputType.PDF, InputType.XLSX]


class BBVA(Entity):
    def parse(self, input_file: Path, context: YnabUnlinkedContext) -> list[Transaction]:
        import datetime as dt

        from ynab_unlinked.exceptions import ParsingError
        from ynab_unlinked.models import Transaction
        from ynab_unlinked.parsers import pdf, xls
        from ynab_unlinked.utils import extract_type

        input_type = extract_type(input_file, valid=VALID_TYPES)
        match input_type:
            case InputType.XLSX | InputType.XLS:
                generator = xls(input_file, read_after_row_like=XLSX_ROW_TO_READ)
                field_reader = self.__extract_fields_from_xlsx_row
            case InputType.PDF:
                generator = pdf(input_file, allow_empty_columns=False, expected_number_of_columns=3)
                field_reader = self.__extract_fields_from_pdf_row
            case (InputType.TXT | InputType.CSV | InputType.HTML) as file_type:
                raise NotImplementedError(
                    f"BBVA does not support input file of type {file_type.value!r}"
                )
            case never:
                assert_never(never)

        transactions = []

        for row in generator:
            parsed_row = field_reader(cast(list[str], row))
            if parsed_row is None:
                raise ParsingError(
                    input_file,
                    "Malformed Transaction Table: Could not extract the date, payee and a mount for one row",
                )

            date, payee, amount = parsed_row

            # It can be that the row does not represent a transaction. Skip it
            try:
                parsed_date = dt.datetime.strptime(date, "%d/%m/%Y").date()
            except ValueError:
                continue

            transactions.append(
                Transaction(
                    date=parsed_date,
                    payee=payee,
                    amount=float(amount.replace("€", "").replace(",", ".")),
                )
            )

        return transactions

    def __extract_fields_from_pdf_row(self, row: list[str]) -> tuple[str, ...] | None:
        # PDFs should have three columns
        # - Date with 2 lines for the date the transaction took place and when it was approved
        # - Concept, used for payee. Sometimes 2 lines including spending category
        # - Amount
        if len(row) != 3:
            return None

        date = row[0].splitlines()[0]
        payee = row[1].splitlines()[0]
        amount = row[2]

        return date, payee, amount

    def __extract_fields_from_xlsx_row(self, row: list[str]) -> tuple[str, ...] | None:
        # XLSX fields are present from 1 to 4 for date, card number, payee and amount
        return str(row[1]), str(row[3]), str(row[4])

    def name(self) -> str:
        return "BBVA Credit Card"



# type: ignore
from factory.base import Factory

from ynab_unlinked.config.models.v2 import CurrencyFormat


class CurrencyFormatFactory(Factory):
    class Meta:
        model = CurrencyFormat

    iso_code = "EUR"
    decimal_digits = 2
    decimal_separator = "."
    symbol_first = False
    group_separator = ","
    currency_symbol = "€"
    display_symbol = True



import datetime as dt
from collections.abc import Generator
from contextlib import ExitStack
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from freezegun import freeze_time
from pytest_mock import MockerFixture
from typer.testing import CliRunner as TyperRunner

from tests.helpers.types import CliRunner
from tests.helpers.ynab_api import YnabClientStub
from ynab_unlinked.config import get_config
from ynab_unlinked.config.core import VERSION_MAPPING
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.formatter import Formatter
from ynab_unlinked.main import app
from ynab_unlinked.utils import split_quoted_string
from ynab_unlinked.ynab_api import Client


@pytest.fixture
def yul(config: str) -> CliRunner:
    def wrapper(*args: str, **kwargs: Any):
        runner = TyperRunner()
        if len(args) == 1 and " " in args[0]:
            # Handle passing all commands as a single string
            args = tuple(split_quoted_string(args[0]))
        result = runner.invoke(app, args=args, **kwargs)
        return result

    return wrapper


@pytest.fixture
def load_entity(ynab_api: YnabClientStub):
    from tests.helpers.load_entity import load_entity

    return load_entity()


@pytest.fixture
def context_obj(config: str):
    config_obj = get_config()
    assert config_obj is not None
    return YnabUnlinkedContext(
        config=config_obj,
        formatter=Formatter(
            date_format=config_obj.budget.date_format,
            currency_format=config_obj.budget.currency_format,
        ),
        extras=None,
    )


@pytest.fixture
def today() -> Generator[dt.datetime]:
    today = dt.datetime(2025, 5, 15)
    with freeze_time("2025-05-15"):
        yield today


@pytest.fixture
def ynab_api(mocker: MockerFixture):
    client_mock = mocker.patch.object(Client, "api")
    stub = YnabClientStub(api_key="someapikey")
    client_mock.side_effect = stub.api
    return stub


@pytest.fixture
def config(request: pytest.FixtureRequest) -> Generator[str]:
    """
    The config version can be requested either by marking the test with

    ```
    @pytest.mark.version("V2")
    ```

    or by passing the version as a string to the fixture request through indirect.

    If the fixture is accessed it yields the version it is pointing to
    """
    # Try to get version from marker first
    version_marker = request.node.get_closest_marker("version")
    if version_marker and version_marker.args:
        version = version_marker.args[0]
    # Fall back to fixture parameter if no marker
    elif hasattr(request, "param"):
        version = request.param
    else:
        pytest.fail(
            "When using the config fixture, either:\n"
            "1. Use a version marker: @pytest.mark.version('V1')\n"
            "2. Pass a version parameter through indirect parameterization"
        )

    config_paths_to_patch = [
        "ynab_unlinked.config.core.config_path",
        *[f"ynab_unlinked.config.models.{v.lower()}.config_path" for v in VERSION_MAPPING],
    ]

    with ExitStack() as stack:
        for module in config_paths_to_patch:

            def side_effect(v: str | None = None) -> Path:
                if not version.startswith("V"):
                    # For any special version, get whatever file is requested
                    # Make sure we do not accept any check for V1
                    if v == "V1":
                        return Path(f"tests/assets/config_{version}/no_config.json")

                    return Path(f"tests/assets/config_{version}/config.json")

                # When loading the path of a module that is not the same as the version being patched
                # set the path to a non-existing file
                if v == version:
                    return Path(f"tests/assets/config_{version}/config.json")

                return Path(f"tests/assets/config_{version}/no_config.json")

            stack.enter_context(patch(module, side_effect=side_effect))
        yield version



from collections.abc import Generator, Sequence
from pathlib import Path
from typing import Any


def xls(
    input_file: Path,
    read_after_row: int = 0,
    read_after_row_like: Sequence[str] | None = None,
    allow_partial_match: bool = False,
) -> Generator[Sequence[str]]:
    """Reads rows from an XLS file, yielding each row as a sequence of strings.

    Skips a specified number of rows or until a matching row is found, then yields subsequent rows.
    Allows for partial or full row matching to determine where to start reading.

    Args:
        input_file: The path to the XLS file to read.
        read_after_row: The number of initial rows to skip before reading.
        read_after_row_like: A sequence of strings to match a row after which reading should begin.
        allow_partial_match: If True, allows partial matching of the row to start reading.

    Returns:
        Generator[Sequence[str]]: A generator yielding each row as a sequence of strings.
    """
    import pyexcel

    kwargs: dict[str, Any] = {"file_name": str(input_file.absolute())}

    read = False

    for idx, entry in enumerate(pyexcel.get_array(**kwargs)):
        if idx < read_after_row:
            continue

        if read_after_row_like is not None and not read:
            if allow_partial_match:
                for after_row_idx, element in enumerate(read_after_row_like):
                    if entry[after_row_idx] != element:
                        break
                else:
                    # We have completed the loop, we are in the right place
                    read = True
            elif entry == read_after_row_like:
                read = True
            continue

        if read:
            yield entry



import datetime as dt
from typing import Annotated

from typer import Context, Option
from ynab import Account, TransactionClearedStatus, TransactionDetail

from ynab_unlinked import app, display
from ynab_unlinked.choices import Choice
from ynab_unlinked.commands.recapp import Reconcile
from ynab_unlinked.config import ConfigV2
from ynab_unlinked.config.constants import TRANSACTION_GRACE_PERIOD_DAYS
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.display import process
from ynab_unlinked.ynab_api import Client


def build_choices(transactions: list[TransactionDetail], accounts: list[Account]) -> list[Choice]:
    accounts_by_id = {acc.id: acc for acc in accounts}
    choices_per_account: dict[str, list[Choice | str]] = {}

    for transaction in transactions:
        choice = Choice(id=f"transaction-{transaction.id}", transaction=transaction)
        forced_selection = (
            False if transaction.cleared is TransactionClearedStatus.UNCLEARED else None
        )
        choice.enable_forced_selected(forced_selection)
        choices_per_account.setdefault(transaction.account_id, []).append(choice)

    return [
        Choice(
            id=f"account-{acc_id}",
            title=account.name,
            choices=choices_per_account.get(acc_id),
            account=account,
        )
        for acc_id, account in accounts_by_id.items()
        if choices_per_account.get(acc_id)
    ]


@app.command()
def reconcile(
    context: Context,
    all: Annotated[
        bool,
        Option(
            "--all",
            "-a",
            help=(
                "Include all transactions, not just those since the last reconciliation. "
                "Use this if some transactions were cleared with a significant delay. "
                "Note: This may take longer to run. "
                "Alternatively, use the --buffer option to include more days before the last reconciliation."
            ),
        ),
    ] = False,
    buffer: Annotated[
        int,
        Option(
            "-b",
            "--buffer",
            help=(
                "Number of days before the last reconciliation to include when checking transactions. "
                "This helps catch any late-cleared transactions."
            ),
            show_default=True,
        ),
    ] = 7,
):
    """Help reconciling your accounts in one go"""

    ctx: YnabUnlinkedContext = context.obj
    config: ConfigV2 = ctx.config

    budget_id = config.budget.id

    last_reconciliation_date = None if all else config.last_reconciliation_date

    if last_reconciliation_date:
        last_reconciliation_date -= dt.timedelta(days=buffer)

    client = Client(api_key=config.api_key)

    with process("Getting transactions from YNAB"):
        transactions_to_reconcile = [
            transaction
            for transaction in client.transactions(
                budget_id=budget_id, since_date=last_reconciliation_date
            )
            if transaction.cleared is not TransactionClearedStatus.RECONCILED
        ]
        accounts = client.accounts(budget_id=budget_id)

    if not transactions_to_reconcile:
        display.success("All accounts are already reconciled!")
        return

    choices = build_choices(transactions_to_reconcile, accounts)
    return_value = Reconcile(config, choices, formatter=ctx.formatter).run()

    if return_value == 2:
        display.info("Nothing to reconcile.\n👋 Bye!")
        return

    if return_value == 1:
        display.info("👋 Bye!")
        return

    selected_transactions = [
        child.transaction
        for choice in choices
        for child in choice.choices
        if child.transaction and child.is_selected
    ]

    for transaction in selected_transactions:
        transaction.cleared = TransactionClearedStatus.RECONCILED

    with process("Updating transactions"):
        client.update_transactions(budget_id=budget_id, transactions=selected_transactions)

    latest_date = max(t.var_date for t in selected_transactions)
    config.last_reconciliation_date = latest_date - dt.timedelta(days=TRANSACTION_GRACE_PERIOD_DAYS)
    config.save()

    display.success("🎉 Reconciliation done!")



import datetime as dt
from typing import Literal, TypedDict, overload

from ynab.api.accounts_api import AccountsApi
from ynab.api.budgets_api import BudgetsApi
from ynab.api.payees_api import PayeesApi
from ynab.api.transactions_api import TransactionsApi
from ynab.api_client import ApiClient
from ynab.configuration import Configuration
from ynab.models.account import Account
from ynab.models.budget_detail import BudgetDetail
from ynab.models.budget_summary import BudgetSummary
from ynab.models.new_transaction import NewTransaction
from ynab.models.patch_transactions_wrapper import PatchTransactionsWrapper
from ynab.models.payee import Payee
from ynab.models.post_transactions_wrapper import PostTransactionsWrapper
from ynab.models.save_transaction_with_id_or_import_id import SaveTransactionWithIdOrImportId
from ynab.models.transaction_detail import TransactionDetail

from ynab_unlinked.models import TransactionWithYnabData


class ApisType(TypedDict):
    budget: type[BudgetsApi]
    accounts: type[AccountsApi]
    transactions: type[TransactionsApi]
    payees: type[PayeesApi]


SupportedApisType = BudgetsApi | AccountsApi | TransactionsApi | PayeesApi
SupportedApisNames = Literal["budget", "accounts", "transactions", "payees"]


class Client:
    def __init__(self, api_key: str):
        self.api_key = api_key
        self.__client = ApiClient(Configuration(access_token=api_key))
        self._apis: ApisType = {
            "budget": BudgetsApi,
            "accounts": AccountsApi,
            "transactions": TransactionsApi,
            "payees": PayeesApi,
        }

    @overload
    def api(self, api_name: Literal["budget"]) -> BudgetsApi: ...

    @overload
    def api(self, api_name: Literal["accounts"]) -> AccountsApi: ...

    @overload
    def api(self, api_name: Literal["transactions"]) -> TransactionsApi: ...

    @overload
    def api(self, api_name: Literal["payees"]) -> PayeesApi: ...

    def api(self, api_name: SupportedApisNames) -> SupportedApisType:
        if (api := self._apis.get(api_name)) is None:
            raise ValueError(f"The api {api_name!r} is not supported")

        return api(self.__client)

    def budgets(self, include_accounts: bool = False) -> list[BudgetSummary]:
        api = self.api("budget")
        response = api.get_budgets(include_accounts=include_accounts)
        return response.data.budgets

    def budget(self, budget_id: str) -> BudgetDetail:
        api = self.api("budget")
        response = api.get_budget_by_id(budget_id=budget_id)
        return response.data.budget

    def accounts(self, budget_id: str) -> list[Account]:
        api = self.api("accounts")
        response = api.get_accounts(budget_id)
        return response.data.accounts

    def transactions(
        self,
        budget_id: str,
        account_id: str | None = None,
        since_date: dt.datetime | dt.date | None = None,
    ) -> list[TransactionDetail]:
        api = self.api("transactions")

        if since_date is not None and isinstance(since_date, dt.datetime):
            since_date = since_date.replace(hour=0, minute=0, second=0, microsecond=0)

        if account_id:
            response = api.get_transactions_by_account(
                budget_id=budget_id,
                account_id=account_id,
                since_date=since_date,
            )
        else:
            response = api.get_transactions(
                budget_id=budget_id,
                since_date=since_date,
            )

        return response.data.transactions

    def payees(self, budget_id: str) -> list[Payee]:
        api = self.api("payees")
        response = api.get_payees(budget_id)
        return response.data.payees

    def create_transactions(
        self,
        budget_id: str,
        account_id: str,
        transactions: list[TransactionWithYnabData],
    ):
        if not transactions:
            return

        api = self.api("transactions")

        transactions_to_create = [
            NewTransaction(
                account_id=account_id,
                date=t.date,
                payee_name=t.payee,
                cleared=t.cleared,
                amount=int(t.amount * 1000),
                approved=False,
                import_id=t.id,
            )
            for t in transactions
        ]

        api.create_transaction(
            budget_id,
            data=PostTransactionsWrapper(transactions=transactions_to_create),
        )

    def update_transactions(self, budget_id: str, transactions: list[TransactionDetail]):
        api = self.api("transactions")

        to_update = [
            SaveTransactionWithIdOrImportId(
                id=t.id,
                account_id=t.account_id,
                cleared=t.cleared,
            )
            for t in transactions
        ]
        api.update_transactions(
            budget_id=budget_id,
            data=PatchTransactionsWrapper(transactions=to_update),
        )



from __future__ import annotations

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import datetime as dt
    from pathlib import Path

    from ynab_unlinked.context_object import YnabUnlinkedContext
    from ynab_unlinked.models import Transaction


# Line that triggers the credit operations
ANCHOR_LINE = "Límite de crédito".encode("cp1252").decode("cp1252")
XLS_DEBIT_LINE = "MOVIMIENTOS DE DEBITO"
TRANSACTION_PATTERN = re.compile(r"^(\d{2}/\d{2})\|(.+?)\|.+?\|(\d+.*EUR)(\([\d*]\))?$")


class SabadellParser:
    def parse(self, input_file: Path, context: YnabUnlinkedContext) -> list[Transaction]:
        from ynab_unlinked.entities import InputType
        from ynab_unlinked.utils import extract_type

        input_type = extract_type(input_file, valid=[InputType.TXT, InputType.XLS])

        match input_type:
            case InputType.TXT:
                return self.__parse_txt(input_file)
            case InputType.XLS:
                return self.__parse_xls(input_file)
            case _:
                # Should never happen because we already checked
                raise RuntimeError(f"Unexpected input type: {input_type}")

    def __parse_txt(self, input_file: Path) -> list[Transaction]:
        from ynab_unlinked.models import Transaction

        lines = input_file.read_text(encoding="cp1252").splitlines()
        start = False
        transactions: list[Transaction] = []
        for line in lines:
            if ANCHOR_LINE in line:
                start = True
                continue

            if not start:
                continue

            if (match := TRANSACTION_PATTERN.match(line)) is not None:
                print(match.groups())
                if len(match.groups()) == 4 and match[4] == "(1)":
                    # Pending transaction
                    continue

                transactions.append(
                    Transaction(
                        date=self.__parse_date(match[1]),
                        payee=self.__parse_payee(match[2]),
                        amount=-self.__parse_amount(match[3]),
                    )
                )

        return transactions

    def __parse_xls(self, input_file: Path) -> list[Transaction]:
        from ynab_unlinked.models import Transaction
        from ynab_unlinked.parsers import xls

        # This is the row after which real transactions appear
        row_trigger = ["FECHA", "CONCEPTO", "LOCALIDAD"]

        transactions = []

        for entry in xls(input_file, read_after_row_like=row_trigger, allow_partial_match=True):
            # If we find debit movements, stop reading
            # Debit movements appear at the end of the file
            if entry[0] == XLS_DEBIT_LINE:
                break

            # The order is date, payee, x, x, value, EUR
            date, payee, amount, currency = entry[0], entry[1], entry[4], entry[5]

            # If currency ends in (1) is a pending transaction, skip
            if currency.endswith("(1)"):
                continue

            # If we cannot parse the date, then continue because we might be in an entry that is not a
            # transaction entry
            try:
                parsed_date = self.__parse_date(date)
            except Exception:
                continue

            transactions.append(
                Transaction(
                    date=parsed_date,
                    payee=self.__parse_payee(payee),
                    amount=-self.__parse_amount(amount),
                )
            )

        return transactions

    def __parse_date(self, raw: str) -> dt.date:
        import datetime as dt

        current_year = dt.date.today().year
        return dt.datetime.strptime(f"{raw}/{current_year}", "%d/%m/%Y").date()

    def __parse_payee(self, raw: str) -> str:
        return raw.title()

    def __parse_amount(self, raw: str) -> float:
        return float(raw.replace("EUR", "").replace(",", "."))

    def name(self) -> str:
        return "sabadell"



import datetime as dt
import json
from collections.abc import Callable
from pathlib import Path

import pytest

from ynab_unlinked.config import Config
from ynab_unlinked.config.constants import TRANSACTION_GRACE_PERIOD_DAYS
from ynab_unlinked.models import Transaction

# This module tests the central logic of the config object. It does not focus on each particular
# version and instead ensures that the logic that needs to be supported is supported propertly


def record_save(output: list[str]) -> Callable:
    def save_output(self, value: str):
        nonlocal output
        output[0] = value
        return

    return save_output


def test_save(config_obj: Config, monkeypatch: pytest.MonkeyPatch):
    config_obj.api_key = "some-other-api-key"  # type: ignore
    output = [""]

    monkeypatch.setattr(Path, "write_text", record_save(output))

    config_obj.save()
    assert output[0] != ""

    content = json.loads(output[0])
    assert content["api_key"] == "some-other-api-key"


def test_update_and_save(config_obj: Config, monkeypatch: pytest.MonkeyPatch):
    trasaction_date = dt.date(2025, 1, 1)
    transaction = Transaction(date=trasaction_date, payee="Acme Store", amount=-12.34)
    output = [""]

    monkeypatch.setattr(Path, "write_text", record_save(output))

    config_obj.update_and_save(transaction, "sabadell")

    assert output[0] != ""

    sabadell = json.loads(output[0])["entities"]["sabadell"]
    assert dt.datetime.strptime(
        sabadell["checkpoint"]["latest_date_processed"], "%Y-%m-%d"
    ).date() == (trasaction_date - dt.timedelta(days=TRANSACTION_GRACE_PERIOD_DAYS))



from ynab.models.transaction_detail import TransactionDetail

from ynab_unlinked.config import ConfigV2
from ynab_unlinked.models import MatchStatus, TransactionWithYnabData
from ynab_unlinked.payee import payee_matches

TIME_WINDOW_MATCH_DAYS = 10


def __match_date(transaction: TransactionWithYnabData, ynab_transaction: TransactionDetail) -> bool:
    return abs((ynab_transaction.var_date - transaction.date).days) <= TIME_WINDOW_MATCH_DAYS


def __match_amount(
    transaction: TransactionWithYnabData, ynab_transaction: TransactionDetail
) -> bool:
    return ynab_transaction.amount / 1000 == transaction.amount


def __match_single_transaction(
    transaction: TransactionWithYnabData,
    ynab_transactions: list[TransactionDetail],
    ynab_matched: set[str],
    reconcile: bool,
    config: ConfigV2,
):
    for ynab_transaction in ynab_transactions:
        if ynab_transaction.id in ynab_matched:
            continue

        date_window = __match_date(transaction, ynab_transaction)
        similar_payee = payee_matches(transaction, config, ynab_transaction)
        same_amount = __match_amount(transaction, ynab_transaction)

        if date_window and same_amount:
            ynab_matched.add(ynab_transaction.id)
            return __finalize_match(transaction, ynab_transaction, reconcile, similar_payee)


def __finalize_match(
    transaction: TransactionWithYnabData,
    ynab_transaction: TransactionDetail,
    reconcile: bool,
    similar_payee: bool,
):
    transaction.ynab_id = ynab_transaction.id
    # Keep track of the status before the match
    transaction.ynab_cleared = ynab_transaction.cleared
    transaction.partial_match = ynab_transaction

    # Update clared status based on the current status from ynab
    transaction.update_cleared_from_ynab(ynab_transaction, reconcile)

    # If we are able to match the payee then we mark them as matched right away
    transaction.match_status = MatchStatus.MATCHED if similar_payee else MatchStatus.PARTIAL_MATCH
    return


def match_transactions(
    transactions: list[TransactionWithYnabData],
    ynab_transactions: list[TransactionDetail],
    reconcile: bool,
    config: ConfigV2,
):
    """Match imported transactions to existing YNAB transactions"""

    # This keep track of ynab transactions already matched
    # The intention is that if a transaction on the same date, payee and amount
    # happens twice, we do not match different imported transaction
    # to a single existing transaction in YNAB
    ynab_matched: set[str] = set()

    for transaction in transactions:
        __match_single_transaction(transaction, ynab_transactions, ynab_matched, reconcile, config)



from collections.abc import Sequence
from pathlib import Path

from rich import box
from rich.prompt import Prompt
from rich.rule import Rule
from rich.style import Style
from rich.table import Column, Table

from ynab_unlinked.config import get_config
from ynab_unlinked.config.models.v2 import Budget, CurrencyFormat
from ynab_unlinked.display import console, process, question
from ynab_unlinked.entities import InputType
from ynab_unlinked.formatter import Formatter
from ynab_unlinked.models import MatchStatus, Transaction, TransactionWithYnabData
from ynab_unlinked.ynab_api.client import Client

MAX_PAST_TRANSACTIONS_SHOWN = 3


def prompt_for_api_key() -> str:
    return question("What is the API Key to connect to YNAB?", password=True)


def extract_type(input_file: Path, valid: Sequence[InputType] | None = None) -> InputType:
    extension = input_file.suffix[1:]

    if extension not in InputType:
        raise LookupError(f"Extension {extension!r} is not supported. File: {input_file}")

    valid_extensions = {v.value for v in valid} if valid else [v.value for v in InputType]
    if extension not in valid_extensions:
        raise AttributeError(
            f"Input file {input_file} does not have a supported extension. Supported: {valid_extensions}"
        )

    return InputType(extension)


def prompt_for_budget(api_key: str | None = None) -> Budget:
    # If no api_key is provided, try to get it from the config
    if api_key is None:
        if (config := get_config()) is None:
            raise RuntimeError("Could not find config and no API key was provided.")
        api_key = config.api_key

    client = Client(api_key)

    with process("Getting budgets..."):
        budgets = client.budgets()

    console().print("Available budgets:")
    console().print(f" - {idx + 1}. {budget.name}" for idx, budget in enumerate(budgets))

    budget_num = Prompt.ask(
        "What budget do you want to use? (By number)",
        choices=[str(i) for i in range(1, len(budgets) + 1)],
        show_choices=False,
        console=console(),
    )
    selected_budget = budgets[int(budget_num) - 1]

    console().print(f"[bold]Selected budget: {selected_budget.name}")

    # Get the full budget to get all the required fields
    budget_details = client.budget(selected_budget.id)

    if budget_details is None:
        raise ValueError(f"Could not find budget with ID {selected_budget.id}")

    if budget_details.currency_format is None:
        raise ValueError(f"Budget {budget_details.name!r} has no currency format")

    if budget_details.date_format is None:
        raise ValueError(f"Budget {budget_details.name!r} has no date format")

    return Budget(
        id=budget_details.id,
        name=budget_details.name,
        date_format=budget_details.date_format.format,
        currency_format=CurrencyFormat(
            iso_code=budget_details.currency_format.iso_code,
            decimal_digits=budget_details.currency_format.decimal_digits,
            decimal_separator=budget_details.currency_format.decimal_separator,
            symbol_first=budget_details.currency_format.symbol_first,
            group_separator=budget_details.currency_format.group_separator,
            currency_symbol=budget_details.currency_format.currency_symbol,
            display_symbol=budget_details.currency_format.display_symbol,
        ),
    )


def display_transaction_table(transactions: list[Transaction], formatter: Formatter):
    columns = [
        Column(header="Date", justify="left", max_width=10),
        Column(header="Payee", justify="left", width=50),
        Column(header="Inflow", justify="right", max_width=15),
        Column(header="Outflow", justify="right", max_width=15),
    ]
    table = Table(
        *columns,
        title="Transactions to process",
        caption=f"Only {MAX_PAST_TRANSACTIONS_SHOWN} processed transactions are shown.",
        box=box.SIMPLE,
    )

    past_counter = 0
    for transaction in transactions:
        style = Style(color="gray37" if transaction.past else "default")

        past_counter += int(transaction.past)
        if past_counter == MAX_PAST_TRANSACTIONS_SHOWN:
            # Stop adding transactions that are past after 5 for clarification
            table.add_row("...", "...", "...", "...")
            break

        amount_str = formatter.format_amount(transaction.amount)
        outflow = amount_str if transaction.amount < 0 else ""
        inflow = amount_str if transaction.amount > 0 else ""

        table.add_row(
            formatter.format_date(transaction.date),
            transaction.payee,
            inflow,
            outflow,
            style=style,
        )

    console().print(table)


def payee_line(transaction: TransactionWithYnabData) -> str:
    if transaction.ynab_payee is not None and transaction.payee == transaction.ynab_payee:
        return transaction.ynab_payee

    return f"{transaction.ynab_payee} [gray37] [Original payee: {transaction.payee}][/gray37]"


def updload_help_message(with_partial_matches=False) -> str:
    main_message = (
        "The table below shows the transactaions to be imported to YNAB. The transactions in the input file "
        "have been matched with existing transactions in YNAB.\n"
        " - The [green]green[/] rows are new transactions to be imported.\n"
    )
    if with_partial_matches:
        main_message += (
            " - The [yellow]yellow[/] rows are transaction to be imported that match in date and amount with\n"
            "   transations that exist in YNAB but for which teh payee name could not be matched.\n"
            "   This is usually because the name from the import file is substantially different any payee "
            "present in YNAB.\n"
            "   If you accept these transactions are valid, we will keep track of this naming for future imports."
        )

    main_message += (
        "The cleared status column shows how the transaction will be loaded to YNAB, not the current "
        "status if the transaction was already in YNAB."
    )

    return main_message


def display_transactions_to_upload(
    transactions: list[TransactionWithYnabData], formatter: Formatter
):
    if not transactions:
        return

    columns = [
        Column(header="Match", justify="center", width=5),
        Column(header="Date", justify="left", max_width=10),
        Column(header="Payee", justify="left", width=70),
        Column(header="Inflow", justify="right", max_width=15),
        Column(header="Outflow", justify="right", max_width=15),
        Column(header="Cleared Status", justify="left", width=15),
    ]
    table = Table(
        *columns,
        title="Recent Transactions",
        caption="Transactions to [cyan bold]update[/] and [bold green]create[/].",
        box=box.SIMPLE,
    )

    partial_matches = False
    for transaction in transactions:
        amount_str = formatter.format_amount(transaction.amount)
        outflow = amount_str if transaction.amount < 0 else ""
        inflow = amount_str if transaction.amount > 0 else ""

        if transaction.needs_creation:
            if transaction.match_status == MatchStatus.PARTIAL_MATCH:
                style = "yellow"
                partial_matches = True
            else:
                style = "green"
        else:
            style = "default"

        table.add_row(
            transaction.match_emoji,
            formatter.format_date(transaction.date),
            payee_line(transaction),
            inflow,
            outflow,
            transaction.cleared_status,
            style=style,
        )

    console().print(Rule("Transactions to be imported"))
    console().print(updload_help_message(partial_matches))
    console().print(table)


def display_partial_matches(transactions: list[TransactionWithYnabData], formatter: Formatter):
    columns = [
        Column(header="Date", justify="left", max_width=10),
        Column(header="Payee", justify="left", width=50),
        Column(header="Inflow", justify="right", max_width=15),
        Column(header="Outflow", justify="right", max_width=15),
        Column(header="Cleared Status", justify="left", width=15),
    ]
    table = Table(
        *columns,
        title="Partial Matches",
        caption=(
            "Each pair of transactions shows the imported transaction (top) \n"
            "and the partial match in YNAB (bottom)."
        ),
        box=box.SIMPLE,
        row_styles=["", "gray70"],
    )

    for transaction in transactions:
        # If we do not need to import it, skip it
        if not transaction.needs_creation:
            continue

        # Skip if no partial match
        if (
            transaction.match_status != MatchStatus.PARTIAL_MATCH
            or transaction.partial_match is None
        ):
            continue

        # Original transaction row
        orig_amount_str = formatter.format_amount(transaction.amount)
        orig_outflow = orig_amount_str if transaction.amount < 0 else ""
        orig_inflow = orig_amount_str if transaction.amount > 0 else ""

        # YNAB transaction row (from partial_match)
        ynab_amount_str = formatter.format_amount_milli(transaction.partial_match.amount)
        ynab_outflow = ynab_amount_str if transaction.partial_match.amount < 0 else ""
        ynab_inflow = ynab_amount_str if transaction.partial_match.amount > 0 else ""

        # Add the pair of rows
        table.add_row(
            formatter.format_date(transaction.date),
            transaction.payee,
            orig_inflow,
            orig_outflow,
            transaction.cleared_status,
        )

        table.add_row(
            formatter.format_date(transaction.partial_match.var_date),
            transaction.partial_match.payee_name or "",
            ynab_inflow,
            ynab_outflow,
            transaction.partial_match.cleared.name.capitalize(),
            end_section=True,
        )

    console().print(table)


def split_quoted_string(input_string: str) -> list[str]:
    """
    Splits a string by spaces, but keeps substrings within
    single quotes, double quotes, or backticks as single tokens.

    Empty quotes are stripped.

    Examples:
        >>> split_quoted_string("this is my name")
        ['this', 'is', 'my', 'name']
        >>> split_quoted_string("I live in 'La guardia'")
        ['I', 'live', 'in', 'La guardia']
    """
    # Regex Explanation:
    # This regular expression is designed to find one of two types of patterns:
    # 1. A quoted string (single, double, or backtick quotes):
    #    - r"'[^']*'"  : Matches content inside single quotes.
    #      '           : Matches a literal single quote.
    #      [^']* : Matches any character that is NOT a single quote, zero or more times.
    #      '           : Matches the closing single quote.
    #    - r'"[^"]*"'  : Matches content inside double quotes. (Similar logic)
    #    - r"`[^`]*`"  : Matches content inside backticks. (Similar logic)
    # 2. A sequence of non-whitespace characters:
    #    - r'\S+'     : Matches one or more non-whitespace characters (for unquoted words).
    #      \S          : Matches any non-whitespace character.
    #      +           : Matches one or more times.

    # The | (OR) operator combines these patterns. The order is crucial:
    # Quoted patterns come first to ensure they are matched as a whole
    # before individual non-whitespace characters inside them are considered.
    import re

    pattern = re.compile(r"'[^']*'|\"[^\"]*\"|`[^`]*`|\S+")

    # re.findall finds all non-overlapping matches of the pattern in the string.
    matches = pattern.findall(input_string)

    # Post-process: remove the surrounding quotes from the matched substrings
    result = []
    for single_match in matches:
        # Check if the match starts and ends with any of the recognized quote types
        if (
            (single_match.startswith("'") and single_match.endswith("'"))
            or (single_match.startswith('"') and single_match.endswith('"'))
            or (single_match.startswith("`") and single_match.endswith("`"))
        ):
            result.append(single_match[1:-1])
        else:
            result.append(single_match)
    return [string for string in result if string]



from .constants import LATEST_VERSION, MAX_CONFIG_VERSION
from .core import get_config
from .models import ConfigV1, ConfigV2
from .types import Config

__all__ = [
    "ConfigV1",
    "ConfigV2",
    "Config",
    "get_config",
    "LATEST_VERSION",
    "MAX_CONFIG_VERSION",
]



from typing import NamedTuple, Protocol


class Version(NamedTuple):
    entity: str
    version: str

    def __ge__(self, value: tuple[str, ...]) -> bool:
        return self.version >= value[1]

    def __le__(self, value: tuple[str, ...]) -> bool:
        return self.version <= value[1]

    def __str__(self) -> str:
        return f"{self.entity}:{self.version}"

    def __repr__(self) -> str:
        return f"'{self.entity}:{self.version}'"


class Versioned(Protocol):
    @staticmethod
    def version() -> Version: ...



from pathlib import Path

from platformdirs import user_config_dir

from .constants import LATEST_VERSION


def v1_config_path() -> Path:
    # This needs to be done especifically for version 1 because it was stored in a different path
    # and was the version before any versioning was implemented
    return Path.home() / ".config/ynab_unlinked/config.json"


def config_path(version: str = LATEST_VERSION.version) -> Path:
    if version == "V1":
        return v1_config_path()

    return Path(user_config_dir("ynab-unlinked", "committhatline")) / "config.json"



from __future__ import annotations

from ynab import Account, TransactionDetail


class Choice:
    def __init__(
        self,
        id: str,
        transaction: TransactionDetail | None = None,
        account: Account | None = None,
        title: str = "",
        choices: list[str | Choice] | None = None,
        selected=False,
        parent: Choice | None = None,
    ):
        self.id = id
        self.title = title
        self._transaction = transaction
        self._account = account
        self.choices: list[Choice] = []
        self.selected = selected
        self._forced_selection: bool | None = None
        self.parent = parent
        if choices is not None:
            n = 0
            for choice in choices:
                if isinstance(choice, str):
                    self.choices.append(Choice(title=choice, id=f"{self.id}-{n}", parent=self))
                    n += 1
                else:
                    choice.parent = self
                    self.choices.append(choice)

    @property
    def transaction(self) -> TransactionDetail:
        if self._transaction is None:
            raise ValueError(
                f"Transaction is needed in choice {self.id!r} but has not been provided."
            )
        return self._transaction

    @property
    def account(self) -> Account:
        if self._account is None:
            raise ValueError(f"Account is needed in choice {self.id!r} but has not been provided.")
        return self._account

    def has_choices(self) -> bool:
        return len(self.choices) > 0

    def select(self):
        """Mark the choice as selected. This implicitely select all child choices."""
        self.selected = True

    def deselect(self):
        """Marks choice as not selected. Child choices keep their value."""
        self.selected = False

    def toggle_selection(self):
        """Toggles the value of the own selected flag."""
        if self.is_forced_selected():
            return
        self.selected = not self.selected

    @property
    def is_selected(self) -> bool:
        """If the choice has a parent that is selected, it is considered selected."""
        if self.is_forced_selected():
            assert self._forced_selection is not None, (
                "Cannot have forced seleciton and not have it set"
            )
            return self._forced_selection

        if self.parent is not None and self.parent.is_selected:
            return True
        return self.selected

    def enable_forced_selected(self, value: bool | None):
        self._forced_selection = value

    def disable_forced_selected(self):
        self._forced_selection = None

    def is_forced_selected(self) -> bool:
        """Whether we need to ignore selected logic and apply any forced selection enforced from the outside"""
        return self._forced_selection is not None

    @property
    def has_selected_choices(self) -> bool:
        """Returns whether or not any of the inner choices is selected"""
        return any(child.is_selected for child in self.choices)

    def to_dict(self) -> dict[str, Choice]:
        """
        Return a dictionary where the key is the id of the Choice and the value is the choice.

        This dictionary flattens out all children choices.
        """
        if not hasattr(self, "__dict"):
            dict_value: dict[str, Choice] = {self.id: self}

            for choice in self.choices:
                dict_value |= choice.to_dict()

            self.__dict = dict_value

        return self.__dict



from __future__ import annotations

import json
from typing import Final

from .constants import LATEST_VERSION
from .migrations import MigrationEngine, Version
from .models import ConfigV1, ConfigV2, DeltaConfigV1ToV2
from .paths import config_path
from .types import Config

VERSION_MAPPING: Final[dict[str, type[Config]]] = {
    "V1": ConfigV1,
    "V2": ConfigV2,
}
LATESST_CONFIG_TYPE = ConfigV2


class ConfigError(ValueError): ...


def config_version() -> Version:
    # Check V1
    # V1 does not have a version in it and was stored in a different path
    if config_path("V1").is_file():
        return Version("Config", "V1")

    latest_config_path = config_path(LATEST_VERSION.version)
    if not latest_config_path.is_file():
        # This can happen when we run the tool for the first time. Return the latest verison
        return LATEST_VERSION

    with open(latest_config_path) as config_file:
        content = json.load(config_file)
        if "version" not in content:
            raise ConfigError(
                "Configuration file malformatted. Run `yul config reset` to reconfigure yul."
            )
        return Version("Config", content["version"])


def get_config() -> LATESST_CONFIG_TYPE | None:
    """Get the latest supported version of the config running any migrations if needed"""
    version = config_version()

    if (current_config := VERSION_MAPPING.get(version.version)) is None:
        raise ConfigError(f"Unsupported config version: {version!r}")

    if not current_config.exists():
        return None

    if current_config is LATESST_CONFIG_TYPE:
        return LATESST_CONFIG_TYPE.load()

    return MigrationEngine("Config", DeltaConfigV1ToV2()).migrate(
        current_config.load(), LATESST_CONFIG_TYPE
    )



from pathlib import Path

import pytest

from ynab_unlinked.entities import InputType
from ynab_unlinked.utils import extract_type


@pytest.mark.parametrize(
    "input_file, expected_type",
    [
        pytest.param("file.txt", InputType.TXT, id="txt"),
        pytest.param("file.csv", InputType.CSV, id="csv"),
        pytest.param("file.html", InputType.HTML, id="html"),
        pytest.param("file.xls", InputType.XLS, id="xls"),
        pytest.param("file.xlsx", InputType.XLSX, id="xlsx"),
        pytest.param("file.pdf", InputType.PDF, id="pdf"),
    ],
)
def test_valid_input_types(input_file: str, expected_type: InputType):
    assert extract_type(Path(input_file)) == expected_type


def test_not_supported_input_type():
    with pytest.raises(LookupError):
        extract_type(Path("file.broken"))


def test_invalid_extension():
    with pytest.raises(AttributeError):
        extract_type(Path("file.pdf"), [InputType.TXT, InputType.CSV])



from __future__ import annotations

import shutil

from ynab_unlinked.config.migrations.base import Delta
from ynab_unlinked.ynab_api import Client

from .v1 import ConfigV1
from .v2 import Budget, ConfigV2, CurrencyFormat


class DeltaConfigV1ToV2(Delta[ConfigV1, ConfigV2]):
    origin = ConfigV1.version()
    destination = ConfigV2.version()

    def on_migrate(self, origin: ConfigV1) -> ConfigV2:
        client = Client(origin.api_key)

        budget_details = client.budget(origin.budget_id)
        if budget_details is None:
            raise ValueError(f"Could not find budget with ID {origin.budget_id}")

        if budget_details.currency_format is None:
            raise ValueError(f"Budget {budget_details.name!r} has no currency format")

        if budget_details.date_format is None:
            raise ValueError(f"Budget {budget_details.name!r} has no date format")

        # Create CurrencyFormat from budget details
        currency_format = CurrencyFormat(
            iso_code=budget_details.currency_format.iso_code,
            decimal_digits=budget_details.currency_format.decimal_digits,
            decimal_separator=budget_details.currency_format.decimal_separator,
            symbol_first=budget_details.currency_format.symbol_first,
            group_separator=budget_details.currency_format.group_separator,
            currency_symbol=budget_details.currency_format.currency_symbol,
            display_symbol=budget_details.currency_format.display_symbol,
        )

        # Create Budget object
        budget = Budget(
            id=budget_details.id,
            name=budget_details.name,
            date_format=budget_details.date_format.format,
            currency_format=currency_format,
        )

        # Create and return ConfigV2 with all fields
        config_v2 = ConfigV2(
            api_key=origin.api_key,
            budget=budget,
            last_reconciliation_date=origin.last_reconciliation_date,
            entities=origin.entities,
            payee_rules=origin.payee_rules,
        )

        # Now we delete v1 and save v2
        # This is special because v1 is stored in a different path than any other versions
        shutil.rmtree(ConfigV1.path().parent)
        config_v2.save()

        return config_v2

    def on_rollback(self, destination: ConfigV2) -> ConfigV1:
        config_v1 = ConfigV1(
            api_key=destination.api_key,
            budget_id=destination.budget.id,
            last_reconciliation_date=destination.last_reconciliation_date,
            entities=destination.entities,
            payee_rules=destination.payee_rules,
        )

        # Ensure we store it on the expected place of v1
        ConfigV1.path().parent.mkdir(parents=True, exist_ok=True)
        config_v1.save()
        # And now delete the other config
        ConfigV2.path().unlink()

        return config_v1



from typing import overload

import unidecode
from rapidfuzz import fuzz
from ynab.models.payee import Payee
from ynab.models.transaction_detail import TransactionDetail

from ynab_unlinked.config import ConfigV2
from ynab_unlinked.models import TransactionWithYnabData
from ynab_unlinked.ynab_api import Client

FUZZY_MATCH_THRESHOLD = 90


def __preprocess_payee(value: str) -> str:
    result = value.lower().replace(" ", "")
    return unidecode.unidecode(result)


@overload
def payee_matches(
    transaction: TransactionWithYnabData,
    config: ConfigV2,
    payee_source: TransactionDetail,
) -> bool: ...


@overload
def payee_matches(
    transaction: TransactionWithYnabData, config: ConfigV2, payee_source: Payee
) -> bool: ...


def payee_matches(
    transaction: TransactionWithYnabData,
    config: ConfigV2,
    payee_source: TransactionDetail | Payee,
) -> bool:
    if isinstance(payee_source, TransactionDetail):
        if payee_source.payee_name is None:
            return False
        payee_name = payee_source.payee_name
    else:
        payee_name = payee_source.name

    if payee_name == transaction.payee:
        return True

    if config.payee_from_fules(transaction.payee) == payee_name:
        return True

    return (
        fuzz.partial_ratio(
            transaction.payee,
            payee_name,
            score_cutoff=FUZZY_MATCH_THRESHOLD,
            processor=__preprocess_payee,
        )
        > 0
    )


def __match_from_payee_list(
    transaction: TransactionWithYnabData, payees: list[Payee], config: ConfigV2
):
    # If we have a partial match, use it
    if transaction.partial_match is not None:
        transaction.ynab_payee = transaction.partial_match.payee_name
        transaction.ynab_payee_id = transaction.partial_match.payee_id
        return

    for p in payees:
        if payee_matches(transaction, config, p):
            transaction.ynab_payee = p.name
            transaction.ynab_payee_id = p.id
            return

    transaction.ynab_payee = transaction.payee


def set_payee_from_ynab(
    transactions: list[TransactionWithYnabData], client: Client, config: ConfigV2
):
    """
    Compare each transaction payee with an existing YNAB payee and set the payee from YNAB if a match is found
    """
    payees = None
    for t in transactions:
        # First check if we have previous naming rules
        if payee := config.payee_from_fules(t.payee):
            t.ynab_payee = payee
            continue

        # Only call once but do not call unless we have not found a match
        if payees is None:
            payees = client.payees(budget_id=config.budget.id)

        __match_from_payee_list(t, payees, config)



import datetime as dt

from pydantic import BaseModel


class Checkpoint(BaseModel):
    latest_date_processed: dt.date
    latest_transaction_hash: int


class EntityConfig(BaseModel):
    account_id: str
    checkpoint: Checkpoint | None = None



from .config_migrations import DeltaConfigV1ToV2
from .shared import Checkpoint, EntityConfig
from .v1 import ConfigV1
from .v2 import ConfigV2

__all__ = ["ConfigV1", "ConfigV2", "DeltaConfigV1ToV2", "Checkpoint", "EntityConfig"]



from enum import StrEnum
from typing import Annotated, assert_never

import typer

from ynab_unlinked.config import ConfigV2
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.display import console, success
from ynab_unlinked.utils import prompt_for_api_key, prompt_for_budget


class ValidKeys(StrEnum):
    BUDGET = "budget"
    API_KEY = "api_key"


config_app = typer.Typer(help="Manage YNAB Unlinked configuration")


@config_app.command(name="set")
def set_command(
    context: typer.Context,
    key: Annotated[ValidKeys, typer.Argument(help="The config key to set", show_default=False)],
):
    """Set configuration options"""
    ctx: YnabUnlinkedContext = context.obj
    config: ConfigV2 = ctx.config

    match key:
        case ValidKeys.API_KEY:
            api_key = prompt_for_api_key()
            config.api_key = api_key
            config.save()
            success("🎉 The API key has been updated")
        case ValidKeys.BUDGET:
            budget = prompt_for_budget(config.api_key)
            config.budget = budget
            config.save()
            success("🎉 The budget has been updated")
        case never:
            assert_never(never)


@config_app.command(name="show")
def show(context: typer.Context):
    ctx: YnabUnlinkedContext = context.obj
    config: ConfigV2 = ctx.config
    console().print(config.model_dump_json(indent=2))



import datetime as dt
from typing import Protocol

from typer.testing import Result

from ynab_unlinked.models import Transaction


class CliRunner(Protocol):
    def __call__(self, *args: str, input: str | None = None) -> Result: ...


class LoadEntityCallback(Protocol):
    def __call__(
        self, current_date: dt.datetime, transactions: list[Transaction] | None = None
    ): ...



import datetime as dt

import pytest

from tests.helpers import assets
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.entities.bbva.bbva import BBVA

pytestmark = [pytest.mark.version("V2"), pytest.mark.usefixtures("config")]


def test_parse_xlsx(context_obj: YnabUnlinkedContext):
    transactions = BBVA().parse(assets.path("bbva/bbva.xlsx"), context_obj)

    assert len(transactions) == 17
    # Transaction 1 is July 11, Netflix.com for 19.99
    assert transactions[0].date == dt.date(2025, 7, 11)
    assert transactions[0].payee == "Netflix.com"
    assert transactions[0].amount == -19.99

    # Transaction 5 is a payment
    assert transactions[4].date == dt.date(2025, 7, 5)
    assert transactions[4].payee == "Recibo mes anterior"
    assert transactions[4].amount == 270.74



import datetime as dt

import pytest

from tests.factories import CurrencyFormatFactory
from ynab_unlinked.formatter import Formatter


@pytest.fixture
def currency_formatter(request: pytest.FixtureRequest):
    return Formatter("DD/MM/YYYY", request.param)


def formatter_currency_parameterization() -> list:
    params = [
        pytest.param(
            CurrencyFormatFactory.build(),
            1234.56,
            "1,234.56€",
            id="default_eur_positive",
        ),
        pytest.param(
            CurrencyFormatFactory.build(),
            -1234.56,
            "-1,234.56€",
            id="default_eur_negative",
        ),
        pytest.param(
            CurrencyFormatFactory.build(),
            0,
            "0.00€",
            id="zero",
        ),
        pytest.param(
            CurrencyFormatFactory.build(symbol_first=True, currency_symbol="$", iso_code="USD"),
            -543.21,
            "-$543.21",
            id="usd_symbol_first_negative",
        ),
        pytest.param(
            CurrencyFormatFactory.build(decimal_separator=",", group_separator="."),
            987654.32,
            "987.654,32€",
            id="eur_different_separators_large_number",
        ),
        pytest.param(
            CurrencyFormatFactory.build(display_symbol=False),
            1234.56,
            "1,234.56",
            id="no_symbol",
        ),
        pytest.param(
            CurrencyFormatFactory.build(decimal_digits=3),
            1234.56,
            "1,234.560€",
            id="3_decimal_digits",
        ),
        pytest.param(
            CurrencyFormatFactory.build(group_separator=""),
            -1234.56,
            "-1234.56€",
            id="no_group_separator_negative",
        ),
        pytest.param(
            CurrencyFormatFactory.build(group_separator=" ", decimal_separator=","),
            1234.56,
            "1 234,56€",
            id="space_group_separator_comma_decimal",
        ),
    ]
    return params


@pytest.mark.parametrize(
    "currency_formatter, amount, expected_output",
    formatter_currency_parameterization(),
    indirect=["currency_formatter"],
)
def test_currency_formatter(currency_formatter: Formatter, amount: float, expected_output: str):
    assert currency_formatter.format_amount(amount) == expected_output


def formatter_milli_currency_parameterization():
    return [
        pytest.param(
            CurrencyFormatFactory.build(decimal_digits=2),
            123450,
            "123.45€",
            id="milli_eur_positive",
        ),
        pytest.param(
            CurrencyFormatFactory.build(decimal_digits=2),
            -123450,
            "-123.45€",
            id="milli_eur_negative",
        ),
        pytest.param(
            CurrencyFormatFactory.build(
                symbol_first=True, currency_symbol="$", iso_code="USD", decimal_digits=2
            ),
            -123450,
            "-$123.45",
            id="milli_usd_negative_symbol_first",
        ),
        pytest.param(
            CurrencyFormatFactory.build(decimal_digits=0),
            123567,
            "124€",
            id="milli_eur_rounding",
        ),
    ]


@pytest.mark.parametrize(
    "currency_formatter, amount, expected",
    formatter_milli_currency_parameterization(),
    indirect=["currency_formatter"],
)
def test_format_milli_currency(currency_formatter: Formatter, amount: int, expected: str):
    assert currency_formatter.format_amount_milli(amount) == expected


@pytest.mark.parametrize(
    "date_format, date_obj, expected_str",
    [
        pytest.param("YYYY/MM/DD", dt.date(2025, 12, 30), "2025/12/30", id="YYYY/MM/DD"),
        pytest.param("YYYY-MM-DD", dt.date(2025, 12, 30), "2025-12-30", id="YYYY-MM-DD"),
        pytest.param("DD-MM-YYYY", dt.date(2025, 12, 30), "30-12-2025", id="DD-MM-YYYY"),
        pytest.param("DD/MM/YYYY", dt.date(2025, 12, 30), "30/12/2025", id="DD/MM/YYYY"),
        pytest.param("DD.MM.YYYY", dt.date(2025, 12, 30), "30.12.2025", id="DD.MM.YYYY"),
        pytest.param("MM/DD/YYYY", dt.date(2025, 12, 30), "12/30/2025", id="MM/DD/YYYY"),
        pytest.param("YYYY.MM.DD", dt.date(2025, 12, 30), "2025.12.30", id="YYYY.MM.DD"),
    ],
)
def test_format_date(date_format: str, date_obj: dt.date, expected_str: str):
    currency_format = CurrencyFormatFactory.build()
    formatter = Formatter(date_format, currency_format)
    assert formatter.format_date(date_obj) == expected_str



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import nullcontext
from functools import reduce
from typing import cast

from ynab_unlinked import display

from .types import Version, Versioned


class Delta[V1: Versioned, V2: Versioned](ABC):
    origin: Version
    destination: Version

    def __init_subclass__(cls) -> None:
        class_name = cls.__name__
        if not hasattr(cls, "origin"):
            raise ValueError(f"Class {class_name} needs to define an 'origin' class variable.")

        if not hasattr(cls, "destination"):
            raise ValueError(f"Class {class_name} needs to define an 'destination' class variable.")

    @abstractmethod
    def on_migrate(self, origin: V1) -> V2: ...

    def migrate(self, origin: V1) -> V2:
        """Method to migrate from the delta origin to the destination"""
        if origin.version() != self.origin:
            error_msg = (
                "The origin version supplied to migrate is not supported by this Delta: "
                f"{origin.version()} != {self.origin}"
            )
            raise ValueError(error_msg)

        return self.on_migrate(origin)

    @abstractmethod
    def on_rollback(self, destination: V2) -> V1: ...

    def rollback(self, destination: V2) -> V1:
        """Method to rollback from the Delta destination to its origin"""
        if destination.version() != self.destination:
            error_msg = (
                "The origin supplied to migrate is not supported by this Delta: "
                f"{destination.version()} != {self.destination}"
            )
            raise ValueError(error_msg)

        return self.on_rollback(destination)


class DeltaRegistry:
    def __init__(self, cls: str):
        self.cls = cls
        self.origins: dict[str, Delta] = {}
        self.destinations: dict[str, Delta] = {}

    def add_delta(self, delta: Delta):
        """Registers delta in the registry"""
        if delta.origin.version in self.origins:
            raise ValueError(
                f"A delta has already been registered for version {delta.origin.version}"
            )
        self.origins[delta.origin.version] = delta
        self.destinations[delta.destination.version] = delta

    def migration_sequence(self, origin: Version, destination: Version) -> Sequence[Delta]:
        """
        Returns a sequence of detlas to be chained to reach the version in `destination`
        from the `origin` version.
        """
        if origin.entity != destination.entity:
            raise ValueError(
                "Cannot find a sequence between different entities: "
                f"[{origin.entity!r}, {destination.entity!r}]"
            )

        if origin == destination:
            return []

        sequence = []
        is_rollback = origin > destination
        current_delta = (
            self.destinations.get(origin.version)
            if is_rollback
            else self.origins.get(origin.version)
        )

        while current_delta:
            sequence.append(current_delta)
            if (
                # During rollback we are moving backwards
                is_rollback
                and current_delta.origin.version == destination.version
                or current_delta.destination.version == destination.version
            ):
                break

            current_delta = (
                self.destinations.get(current_delta.origin.version)
                if is_rollback
                else self.origins.get(current_delta.destination.version)
            )
        else:
            # Reach this when while ends naturally, i.e. no delta found for the next migration
            raise ValueError(
                f"Cannot find a migration sequence matching {origin.version} -> {destination.version}."
            )

        return sequence


class MigrationEngine:
    _deltas: dict[str, DeltaRegistry] = {}

    def __init__(self, cls: str, *deltas: Delta):
        if cls in self.__class__._deltas:
            raise ValueError(f"The class {cls!r} has already been registered for migrations")

        registry = DeltaRegistry(cls)
        for delta in deltas:
            registry.add_delta(delta)

        self.__class__._deltas[cls] = registry

    def __prepare_sequence(self, origin: Version, destination: Version) -> Sequence[Delta]:
        origin_entity, _ = origin
        destination_entity, _ = destination

        if origin_entity != destination_entity:
            raise ValueError(
                f"Cannot handle migrations on diferent entities: ({origin_entity!r} -> {destination_entity!r})"
            )

        if registry := self._deltas.get(origin_entity):
            return registry.migration_sequence(origin=origin, destination=destination)

        raise ValueError(f"No migrations have been registered for {origin_entity!r}")

    def migrate[V: Versioned](self, origin: Versioned, to_type: type[V]) -> V:
        deltas: Sequence[Delta[Versioned, V]] = self.__prepare_sequence(
            origin.version(), to_type.version()
        )

        process_context = display.process if deltas else lambda *args, **kwargs: nullcontext()

        with process_context(
            f"Migrating {origin.version()} to {to_type.version()}", "Migration done!"
        ):
            # Use of cast because migrate does not produce the same type as output as its input and reduce expects that
            result = cast(V, reduce(lambda result, delta: delta.migrate(result), deltas, origin))

        if not isinstance(result, to_type):
            raise TypeError(
                "Unexpected type mismatch. The result of the migration "
                f"returned {type(result)!r} but {to_type!r} was expected"
            )

        return result

    def rollback[V: Versioned](self, destination: Versioned, to_type: type[V]) -> V:
        deltas: Sequence[Delta[Versioned, V]] = self.__prepare_sequence(
            destination.version(), to_type.version()
        )

        # Ignore typing for a similar reason as in migrate but with the complication that
        # now the types to be returned are inverted
        result: V = reduce(
            lambda result, delta: delta.rollback(result),  # type: ignore
            deltas,
            destination,
        )

        return result



import datetime as dt
from dataclasses import dataclass
from enum import Enum
from hashlib import sha256
from typing import assert_never

from ynab.models.transaction_cleared_status import TransactionClearedStatus
from ynab.models.transaction_detail import TransactionDetail


class MatchStatus(Enum):
    MATCHED = "matched"
    UNMATCHED = "unmatched"
    PARTIAL_MATCH = "partial_match"


@dataclass
class Transaction:
    """Represents a transaction imported from a file by a given entity"""

    date: dt.date
    payee: str
    amount: float

    def __post_init__(self):
        self.past = False
        self.counter = 0

    @property
    def pretty_payee(self) -> str:
        return self.payee if len(self.payee) < 15 else f"{self.payee[:15]}..."

    def __hash__(self) -> int:
        return hash(f"{self.date:%m-%d-%Y}{self.payee}{self.amount}")

    @property
    def id(self) -> str:
        return sha256(
            f"{self.date:%m-%d-%Y}{self.payee}{self.amount}{self.counter}".encode()
        ).hexdigest()[:30]

    @property
    def inflow(self) -> float | None:
        return self.amount if self.amount > 0 else None

    @property
    def outflow(self) -> float | None:
        return self.amount if self.amount < 0 else None

    def __repr__(self) -> str:
        return (
            f"Transaction(date={self.date}, payee={self.payee},"
            f"amount={self.amount}, past={self.past}, counter={self.counter})"
        )


class TransactionWithYnabData(Transaction):
    def __init__(self, transaction: Transaction):
        super().__init__(
            date=transaction.date,
            payee=transaction.payee,
            amount=transaction.amount,
        )
        self.match_status: MatchStatus = MatchStatus.UNMATCHED
        self.partial_match: TransactionDetail | None = None
        self.ynab_id: str | None = None
        self.ynab_payee_id: str | None = None
        self.ynab_payee: str | None = transaction.payee
        # All transactions are cleared by default because we get them from an entity export
        self.cleared: TransactionClearedStatus = TransactionClearedStatus.CLEARED
        self.ynab_cleared: TransactionClearedStatus | None = None

    def __repr__(self) -> str:
        return (
            f"Transaction(date={self.date}, payee={self.payee!r}, "
            f"amount={self.amount}, past={self.past}, counter={self.counter}, "
            f"match_status={self.match_status}, partial_match={self.partial_match!r}, "
            f"ynab_id={self.ynab_id!r}, ynab_payee_id={self.ynab_payee_id!r}, "
            f"ynab_payee={self.ynab_payee!r}, cleared={self.cleared}, "
            f"ynab_cleared={self.ynab_cleared})"
        )

    @property
    def needs_creation(self) -> bool:
        match_uncleared = (
            self.partial_match is not None
            and self.partial_match.cleared is TransactionClearedStatus.UNCLEARED
        )
        is_unmatched = self.match_status is MatchStatus.UNMATCHED

        return is_unmatched or match_uncleared

    @property
    def cleared_status(self) -> str:
        """
        Generate cleared string representation.

        If the transaction does not need to be created and it has an internal `ynab_cleared`
        status set, the representation of `ynab_cleared` is shown instead. Otherwise,
        the representation of `cleared` is shown instead.
        """
        cleared = (
            self.ynab_cleared
            if not self.needs_creation and self.ynab_cleared is not None
            else self.cleared
        )
        return self.cleared_str(cleared)

    @property
    def ynab_cleared_status(self) -> str:
        return self.cleared_str(self.ynab_cleared) if self.ynab_cleared else ""

    @property
    def match_emoji(self) -> str:
        match self.match_status:
            case MatchStatus.MATCHED:
                return "🔗"
            case MatchStatus.PARTIAL_MATCH:
                assert self.partial_match is not None, (
                    "Cannot have a partial match without a transaction"
                )

                if self.partial_match.cleared is TransactionClearedStatus.UNCLEARED:
                    return "🔍"
                return "🔗"
            case _:
                return ""

    @staticmethod
    def cleared_str(cleared: TransactionClearedStatus) -> str:
        match cleared:
            case TransactionClearedStatus.RECONCILED:
                return "🔒 Reconciled"
            case TransactionClearedStatus.CLEARED:
                return "✅ Cleared"
            case TransactionClearedStatus.UNCLEARED:
                return "Uncleared"
            case _ as never:
                assert_never(never)

    def reset_matching(self):
        """Reset the matching status to UNMATCHED and the payee to the original one"""
        self.match_status = MatchStatus.UNMATCHED
        self.partial_match = None
        self.ynab_payee = self.payee
        self.ynab_payee_id = None

    def update_cleared_from_ynab(self, ynab_transaction: TransactionDetail, reconcile: bool):
        if ynab_transaction.cleared is TransactionClearedStatus.RECONCILED or reconcile:
            self.cleared = TransactionClearedStatus.RECONCILED
        elif ynab_transaction.cleared is TransactionClearedStatus.UNCLEARED:
            self.cleared = TransactionClearedStatus.CLEARED



from unittest.mock import MagicMock

from ynab_unlinked.ynab_api.client import Client, SupportedApisNames


class YnabClientStub(Client):
    def __init__(self, *args, **kwargs):
        self.registry: dict[SupportedApisNames, MagicMock] = {}
        super().__init__(*args, **kwargs)

    def api(self, api_name: SupportedApisNames) -> MagicMock:
        registered_mock = self.registry.get(api_name)

        if registered_mock is None:
            registered_mock = MagicMock(spec=self._apis.get(api_name))
            self.registry[api_name] = registered_mock

        return registered_mock

    def budget(self) -> MagicMock:
        return self.registry.get("budget")

    def accounts(self) -> MagicMock:
        return self.registry.get("accounts")

    def transactions(self) -> MagicMock:
        return self.registry.get("transactions")

    def payees(self) -> MagicMock:
        return self.registry.get("payees")



from collections.abc import Generator
from dataclasses import dataclass

import pytest

from ynab_unlinked.config.migrations import Delta, DeltaRegistry, MigrationEngine, Version


@dataclass
class UserV1:
    name: str
    age: int

    @staticmethod
    def version() -> Version:
        return Version("User", "V1")


@dataclass
class UserV2:
    name: str
    age: int
    email: str

    @staticmethod
    def version() -> Version:
        return Version("User", "V2")


@dataclass
class UserV3:
    name: str
    age: int
    email: str
    address: str | None

    @staticmethod
    def version() -> Version:
        return Version("User", "V3")


@dataclass
class Other:
    prop: int

    @staticmethod
    def version() -> Version:
        return Version("Other", "V1")


class DeltaUserV1V2(Delta[UserV1, UserV2]):
    origin = UserV1.version()
    destination = UserV2.version()

    def on_migrate(self, origin: UserV1) -> UserV2:
        return UserV2(name=origin.name, age=origin.age, email="default@email.com")

    def on_rollback(self, destination: UserV2) -> UserV1:
        return UserV1(
            name=destination.name,
            age=destination.age,
        )


class DeltaUserV2V3(Delta[UserV2, UserV3]):
    origin = UserV2.version()
    destination = UserV3.version()

    def on_migrate(self, origin: UserV2) -> UserV3:
        return UserV3(
            name=origin.name,
            age=origin.age,
            email=origin.email,
            address=None,
        )

    def on_rollback(self, destination: UserV3) -> UserV2:
        return UserV2(
            name=destination.name,
            age=destination.age,
            email=destination.email,
        )


@pytest.fixture
def v1() -> UserV1:
    return UserV1(name="John", age=24)


@pytest.fixture
def v2() -> UserV2:
    return UserV2(name="John", age=24, email="john@domain.com")


@pytest.fixture
def v3() -> UserV3:
    return UserV3(name="John", age=24, email="john@domain.com", address="My Street, My City")


@pytest.fixture
def registry() -> DeltaRegistry:
    registry = DeltaRegistry("User")

    registry.add_delta(DeltaUserV1V2())
    registry.add_delta(DeltaUserV2V3())

    return registry


@pytest.fixture
def engine() -> Generator[MigrationEngine]:
    yield MigrationEngine("User", DeltaUserV1V2(), DeltaUserV2V3())

    MigrationEngine._deltas = {}


def test_version():
    assert Version("User", "V2") > Version("User", "V1")
    assert Version("User", "V2") >= Version("User", "V1")
    assert Version("User", "V2") >= Version("User", "V2")
    assert Version("User", "V2") == Version("User", "V2")
    assert Version("User", "V2") <= Version("User", "V2")
    assert Version("User", "V1") <= Version("User", "V2")
    assert Version("User", "V1") < Version("User", "V2")


def test_delta_migrate(v1: UserV1):
    delta = DeltaUserV1V2()

    v2 = delta.migrate(v1)

    assert v2.name == v1.name
    assert v2.age == v1.age
    assert v2.email == "default@email.com"


def test_delta_rollback(v2: UserV2):
    delta = DeltaUserV1V2()

    v1 = delta.rollback(v2)

    assert v2.name == v1.name
    assert v2.age == v1.age
    assert not hasattr(v1, "email")


def test_delta_cannot_migrate_unknown_version():
    delta = DeltaUserV1V2()

    with pytest.raises(ValueError, match="not supported by this Delta"):
        delta.migrate(Other(prop=1))  # type: ignore


def test_delta_cannot_rollback_unknown_version():
    delta = DeltaUserV1V2()

    with pytest.raises(ValueError, match="not supported by this Delta"):
        delta.rollback(Other(prop=1))  # type: ignore


def test_registry_cannot_register_multiple_delta_per_version():
    registry = DeltaRegistry("User")
    registry.add_delta(DeltaUserV1V2())

    with pytest.raises(ValueError, match="has already been registered for version V1"):
        registry.add_delta(DeltaUserV1V2())


def test_delta_sequence_includes_needed_delta(registry: DeltaRegistry):
    sequence = registry.migration_sequence(UserV1.version(), UserV2.version())

    assert len(sequence) == 1
    assert isinstance(sequence[0], DeltaUserV1V2)


def test_delta_sequence_includes_needed_deltas(registry: DeltaRegistry):
    sequence = registry.migration_sequence(UserV1.version(), UserV3.version())

    assert len(sequence) == 2
    assert isinstance(sequence[0], DeltaUserV1V2)
    assert isinstance(sequence[1], DeltaUserV2V3)


def test_delta_sequence_for_rollbacks(registry: DeltaRegistry):
    sequence = registry.migration_sequence(UserV3.version(), UserV1.version())

    assert len(sequence) == 2
    assert isinstance(sequence[0], DeltaUserV2V3)
    assert isinstance(sequence[1], DeltaUserV1V2)


def test_delta_sequence_error_with_wrong_entity(registry: DeltaRegistry):
    with pytest.raises(ValueError, match="Cannot find a sequence between different entities"):
        registry.migration_sequence(UserV1.version(), Other.version())


def test_delta_sequence_error_without_registered_delta():
    registry = DeltaRegistry("User")
    registry.add_delta(DeltaUserV1V2())

    with pytest.raises(ValueError, match="Cannot find a migration sequence matching V1 -> V3"):
        registry.migration_sequence(UserV1.version(), UserV3.version())


def test_delta_sequence_empty_for_no_change(registry: DeltaRegistry):
    sequence = registry.migration_sequence(UserV1.version(), UserV1.version())

    assert len(sequence) == 0


def test_cannot_register_twice_same_class():
    MigrationEngine("Something")

    with pytest.raises(ValueError, match="has already been registered"):
        MigrationEngine("Something")

    MigrationEngine._deltas = {}


def test_migration_through_engine(v1: UserV1, engine: MigrationEngine):
    v2 = engine.migrate(v1, UserV2)

    assert v2.name == v1.name
    assert v2.age == v1.age
    assert v2.email == "default@email.com"


def test_migration_through_engine_full_sequence(v1: UserV1, engine: MigrationEngine):
    v2 = engine.migrate(v1, UserV3)

    assert v2.name == v1.name
    assert v2.age == v1.age
    assert v2.email == "default@email.com"
    assert v2.address is None


def test_rollbacks_through_engine(v2: UserV2, engine: MigrationEngine):
    v1 = engine.rollback(v2, UserV1)

    assert v2.name == v1.name
    assert v2.age == v1.age
    assert not hasattr(v1, "email")


def test_rollback_through_engine_full_sequence(v3: UserV3, engine: MigrationEngine):
    v1 = engine.rollback(v3, UserV1)

    assert v1.name == v3.name
    assert v1.age == v3.age
    assert not hasattr(v1, "email")
    assert not hasattr(v1, "address")


def test_migration_fails_with_different_entities(v3: UserV3, engine: MigrationEngine):
    with pytest.raises(ValueError, match="Cannot handle migrations on diferent entities"):
        engine.migrate(v3, Other)


def test_rollback_fails_with_different_entities(v3: UserV3, engine: MigrationEngine):
    with pytest.raises(ValueError, match="Cannot handle migrations on diferent entities"):
        engine.rollback(v3, Other)


def test_migration_fails_for_unregistered_entity(engine: MigrationEngine):
    with pytest.raises(ValueError, match="No migrations have been registered for 'Other'"):
        engine.migrate(Other(prop=1), Other)


def test_rollback_fails_for_unregistered_entity(engine: MigrationEngine):
    with pytest.raises(ValueError, match="No migrations have been registered for 'Other'"):
        engine.rollback(Other(prop=1), Other)



from __future__ import annotations

from functools import partial

from textual.app import App
from textual.containers import Container, Grid, Horizontal, VerticalScroll
from textual.message import Message
from textual.screen import ModalScreen
from textual.widgets import Button, DataTable, Footer, Header, Label, Static, Switch
from ynab import TransactionClearedStatus

from ynab_unlinked.choices import Choice
from ynab_unlinked.config import Config
from ynab_unlinked.formatter import Formatter
from ynab_unlinked.models import TransactionWithYnabData

POSITIVE_COLOR = "lightgreen"
NEGATIVE_COLOR = "lightcoral"


class AccountTable(Container):
    class IncludeUncleared(Message):
        def __init__(self, value: bool):
            self.value = value
            super().__init__()

    def __init__(self, choice: Choice, formatter: Formatter):
        self.formatter = formatter
        self.choice = choice
        self.include_uncleared = False
        # By default lets mark the account for reconciliation
        self.choice.select()
        super().__init__()

    def compose(self):
        with Container(classes="account-container"):
            with Horizontal(classes="account-table-header"):
                yield Label(self.choice.title, classes="account-title")
                yield Label(self._account_balances())
                with Horizontal(id="switch-container"):
                    yield Label("Reconcile")
                    yield Switch(self.choice.is_selected, id=f"reconcile-{self.choice.id}")
            yield self.choice_to_table()

    def _account_balances(self) -> str:
        account = self.choice.account
        balance_str = partial(
            self.formatter.format_amount,
            positive_style=POSITIVE_COLOR,
            negative_style=NEGATIVE_COLOR,
        )

        cleared_str = balance_str(account.cleared_balance / 1000)
        uncleared_str = balance_str(account.uncleared_balance / 1000)
        balance_str = balance_str(account.balance / 1000)

        return (
            f"[dim]Cleared:[/dim] {cleared_str} + "
            f"[dim]Uncleared:[/dim] {uncleared_str} "
            f"[dim]= Balance[/dim] {balance_str}"
        )

    def _cleared_status(self, choice: Choice) -> str:
        status = choice.transaction.cleared

        if choice.is_selected:
            status = TransactionClearedStatus.RECONCILED

        return TransactionWithYnabData.cleared_str(status)

    def choice_to_table(self) -> DataTable:
        table = DataTable(cursor_type="row")

        table.add_column("Date")
        table.add_column("Payee")
        table.add_column("Inflow")
        table.add_column("Outflow")
        table.add_column("Cleared Status")

        for choice in self.choice.choices:
            t = choice.transaction
            amount = t.amount / 1000
            table.add_row(
                self.formatter.format_date(t.var_date),
                t.payee_name,
                self.formatter.format_amount(amount) if amount > 0 else "",
                self.formatter.format_amount(amount) if amount < 0 else "",
                self._cleared_status(choice),
                key=choice.id,
            )

        self.log.info(table)
        return table

    def on_switch_changed(self, event: Switch.Changed):
        if event.switch.value:
            self.choice.select()
        else:
            self.choice.deselect()
        self.refresh(recompose=True)

    def on_data_table_row_selected(self, event: DataTable.RowSelected):
        choice_id = event.row_key.value
        if choice_id is None:
            return

        child = self.choice.to_dict().get(choice_id)
        if child is None:
            return

        child.toggle_selection()
        self.refresh(recompose=True)

    def refresh_forced_selection(self, value: bool):
        for choice in self.choice.choices:
            if value:
                # Ensure uncleared are allowed to be selected
                choice.disable_forced_selected()
            else:
                # When uncleared, a transaction must always be unslected
                forced_selection = (
                    False
                    if choice.transaction.cleared is TransactionClearedStatus.UNCLEARED
                    else None
                )
                choice.enable_forced_selected(forced_selection)

        self.refresh(recompose=True)


class ReconcileModal(ModalScreen[int]):
    def __init__(self, choices: list[Choice]):
        super().__init__()
        self.choices = choices

    def compose(self):
        accounts_lines = []
        for choice in self.choices:
            n_child_selected = sum(child.is_selected for child in choice.choices)
            counter_str = (
                "All transactions"
                if n_child_selected == len(choice.choices)
                else f"{n_child_selected} transactions"
            )
            accounts_lines.append(f"- {choice.title} ({counter_str})")

        with Grid(id="reconcile-modal"):
            yield Label("Do you want to reconcile the following accounts?", id="message")
            yield Static("\n".join(accounts_lines), id="accounts-list")
            yield Button("No", id="button-no", variant="error")
            yield Button("Yes", id="button-yes", variant="success")

    def on_button_pressed(self, event: Button.Pressed):
        if event.button.id == "button-no":
            self.dismiss(1)
        else:
            self.dismiss(0)


class Reconcile(App[int]):
    TITLE = "YNAB Unlinked"
    SUB_TITLE = "Reconcile"
    BINDINGS = [("ctrl+q", "quit", "Quit")]

    CSS_PATH = "app.tcss"

    def __init__(self, config: Config, choices: list[Choice], formatter: Formatter):
        self.config = config
        self.choices = choices
        self.formatter = formatter
        super().__init__(watch_css=True)

    def compose(self):
        yield Header()
        yield Static(
            "These are the accounts to be reconciled. "
            "You can mark the entire account for reconciliation or just some of the transactions.\n\n"
            "Marking an account to be reconciled will mark all transactions within it as reconciled as well. "
            "Only uncleared transactions will be left un-marked.\nIf you want to mark uncleared transactions as well, "
            "you can enable the switch 'Include Uncleared'.",
            id="main-description",
        )
        with VerticalScroll(id="main-container"):
            for choice in self.choices:
                yield AccountTable(choice, formatter=self.formatter)
        with Horizontal(id="button-row"):
            yield Label("Include Uncleared")
            yield Switch(False, id="uncleared-switch")
            yield Button("Cancel", variant="default", id="cancel-button")
            yield Button("Reconcile", variant="success", id="reconcile-button")
        yield Footer()

    def on_button_pressed(self, event: Button.Pressed):
        button_id = event.button.id

        if button_id == "cancel-button":
            self.exit(1)

        if button_id == "reconcile-button":
            if any(choice.is_selected or choice.has_selected_choices for choice in self.choices):

                def when_dismissed(value: int | None):
                    if value == 1:
                        return
                    self.exit(value if value is not None else 0)

                self.push_screen(ReconcileModal(self.choices), when_dismissed)
            else:
                self.exit(2)

    def on_switch_changed(self, event: Switch.Changed):
        if event.switch.id and event.switch.id == "uncleared-switch":
            for table in self.query(AccountTable):
                table.refresh_forced_selection(event.switch.value)



import datetime as dt

from ynab_unlinked.config.models.v2 import CurrencyFormat


class Formatter:
    def __init__(self, date_format: str, currency_format: CurrencyFormat):
        self.date_format = date_format
        self.py_date_format = (
            date_format.replace("YYYY", "%Y").replace("MM", "%m").replace("DD", "%d")
        )
        self.currency_format = currency_format

    def format_date(self, date: dt.date) -> str:
        return date.strftime(self.py_date_format)

    def format_amount(self, amount: float, positive_style="", negative_style="") -> str:
        sign = "-" if amount < 0 else ""
        style = positive_style if amount > 0 else negative_style
        end_style = ""
        if style:
            end_style = f"[/{style}]"
            style = f"[{style}]"

        amount = abs(amount)
        amount_str = (
            f"{amount:,.{self.currency_format.decimal_digits}f}".replace(",", "G")
            .replace(".", "D")
            .replace("G", self.currency_format.group_separator)
            .replace("D", self.currency_format.decimal_separator)
        )
        if not self.currency_format.display_symbol:
            return f"{sign}{amount_str}"

        if self.currency_format.symbol_first:
            return f"{sign}{self.currency_format.currency_symbol}{amount_str}"
        return f"{style}{sign}{amount_str}{self.currency_format.currency_symbol}{end_style}"

    def format_amount_milli(self, milli_amount: int) -> str:
        amount = milli_amount / 1000.0
        return self.format_amount(amount)



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from ynab.models.budget_detail import BudgetDetail
from ynab.models.currency_format import CurrencyFormat
from ynab.models.date_format import DateFormat

from ynab_unlinked.config import MAX_CONFIG_VERSION
from ynab_unlinked.config.core import VERSION_MAPPING
from ynab_unlinked.config.migrations.base import MigrationEngine, Version
from ynab_unlinked.ynab_api import Client

if TYPE_CHECKING:
    # No need to import really
    from pytest_mock import MockerFixture

ALL_VERSION_PARAMS = [f"V{i}" for i in range(1, MAX_CONFIG_VERSION + 1)]


@dataclass
class UnlinkMock:
    unlink: MagicMock
    rmtree: MagicMock


@pytest.fixture
def unlink(mocker: MockerFixture):
    # Ensure we are not really deleting files
    unlink_mock = mocker.patch.object(Path, "unlink")
    rmtree_mock = mocker.patch("ynab_unlinked.config.models.config_migrations.shutil.rmtree")

    yield UnlinkMock(unlink_mock, rmtree_mock)


@pytest.fixture(autouse=True)
def ynab_client_mock(mocker: MockerFixture):
    budget_patch = mocker.patch.object(Client, "budget")
    budget_patch.return_value = BudgetDetail(
        id="budget_id",
        name="My Budget",
        date_format=DateFormat(format="DD/MM/YYYY"),
        currency_format=CurrencyFormat(
            iso_code="EUR",
            decimal_digits=2,
            decimal_separator=".",
            symbol_first=False,
            group_separator=",",
            currency_symbol="€",
            display_symbol=True,
            example_format="€1,234.56",
        ),
    )
    yield


@contextmanager
def mock_config_paths(origin: Version, destinoation: Version):
    # Allows mocking the versions to be returned by both config versions
    def path_and_method(v: Version):
        if v.version == "V1":
            return "ynab_unlinked.config.paths.v1_config_path"
        else:
            return f"ynab_unlinked.config.models.{v.version.lower()}.config_path"

    with (
        patch(path_and_method(origin)) as origin_path_patch,
        patch(path_and_method(destinoation)) as destination_path_patch,
    ):
        origin_path_patch.return_value = Path(f"tests/assets/config_{origin.version}/config.json")
        destination_path_patch.return_value = Path(
            f"tests/assets/config_{destinoation.version}/config.json"
        )
        yield


def all_migrations_params(rollbback=False):
    result = []
    for vid in range(1, MAX_CONFIG_VERSION + 1):
        v = Version("Config", f"V{vid}")
        for wid in range(1, MAX_CONFIG_VERSION + 1):
            w = Version("Config", f"V{wid}")

            param = pytest.param(v, w, id=f"{v.version} -> {w.version}")
            if rollbback and v > w or not rollbback and v < w:
                result.append(param)
    return result


@pytest.mark.parametrize("origin, destination", all_migrations_params())
def test_migrations_on_migrate(
    origin: Version, destination: Version, unlink: UnlinkMock, migration_engine: MigrationEngine
):
    with mock_config_paths(origin, destination):
        origin_class = VERSION_MAPPING.get(origin.version)
        destination_class = VERSION_MAPPING.get(destination.version)

        assert origin_class is not None, f"There is no mapping for version {origin.version}"
        assert destination_class is not None, f"There is no mapping for version {origin.version}"

        origin_config = origin_class.load()
        destination_config = destination_class.load()

        migrated_config = migration_engine.migrate(origin_config, destination_class)

        assert migrated_config == destination_config

        should_unlink = origin.version == "V1"
        # sourcery skip: no-conditionals-in-tests
        if should_unlink:
            unlink.rmtree.assert_called_once()


@pytest.mark.parametrize("origin, destination", all_migrations_params(True))
def test_migrations_on_rollback(
    origin: Version,
    destination: Version,
    unlink: UnlinkMock,
    migration_engine: MigrationEngine,
):
    with mock_config_paths(origin, destination):
        origin_class = VERSION_MAPPING.get(origin.version)
        destination_class = VERSION_MAPPING.get(destination.version)

        assert origin_class is not None, f"There is no mapping for version {origin.version}"
        assert destination_class is not None, f"There is no mapping for version {origin.version}"

        origin_config = origin_class.load()
        destination_config = destination_class.load()

        migrated_config = migration_engine.rollback(origin_config, destination_class)

        assert migrated_config == destination_config

        should_unlink = origin.version == "V1"
        # sourcery skip: no-conditionals-in-tests
        if should_unlink:
            unlink.unlink.assert_called_once()



import pytest

from ynab_unlinked.config import MAX_CONFIG_VERSION, Config
from ynab_unlinked.config.core import VERSION_MAPPING
from ynab_unlinked.config.migrations.base import MigrationEngine
from ynab_unlinked.config.models import DeltaConfigV1ToV2


@pytest.fixture
def config_obj(config: str) -> Config:
    return VERSION_MAPPING[config].load()


def pytest_generate_tests(metafunc: pytest.Metafunc):
    # This will generate tests for each test that requires config_obj for all possible versions available of config
    if "config_obj" in metafunc.fixturenames:
        metafunc.parametrize(
            "config",
            [f"V{i}" for i in range(1, MAX_CONFIG_VERSION + 1)],
            indirect=True,
            ids=[f"ConfigV{i}" for i in range(1, MAX_CONFIG_VERSION + 1)],
        )


@pytest.fixture(scope="session")
def migration_engine() -> MigrationEngine:
    return MigrationEngine("Config", DeltaConfigV1ToV2())



import datetime as dt
from dataclasses import dataclass
from pathlib import Path

import typer

from ynab_unlinked.commands import load
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.entities import Entity
from ynab_unlinked.models import Transaction
from ynab_unlinked.process import process_transactions
from ynab_unlinked.utils import MAX_PAST_TRANSACTIONS_SHOWN

from .types import LoadEntityCallback


@dataclass
class StubEntity(Entity):
    transactions: list[Transaction]

    def parse(self, input_file: Path, context: YnabUnlinkedContext) -> list[Transaction]:
        return self.transactions

    def name(self) -> str:
        return "test"


def load_entity() -> LoadEntityCallback:
    """
    Load a stub entity that always returns a given set of transactions for testing.

    This entity is avialable through the load command in the test suit and is associated
    with the TestAccountID account id.
    """

    def callback(current_date: dt.datetime, transactions: list[Transaction] | None = None):
        if transactions is None:
            transactions = [
                Transaction(current_date - dt.timedelta(1), "Test Payee 1", 10.00),
                Transaction(current_date - dt.timedelta(2), "Test Payee 2", -10.00),
                Transaction(
                    current_date - dt.timedelta(MAX_PAST_TRANSACTIONS_SHOWN + 1),
                    "Test Payee 3",
                    -0.15,
                ),
            ]

        def command(context: typer.Context):
            ctx: YnabUnlinkedContext = context.obj

            process_transactions(
                StubEntity(transactions=transactions), input_file=Path(), context=ctx
            )

        load.command(name="test")(command)

    return callback



import pytest

from ynab_unlinked.utils import split_quoted_string


@pytest.mark.parametrize(
    "input_string, expected_output",
    [
        # Cases without quotes
        pytest.param("this is my name", ["this", "is", "my", "name"], id="no_quotes_basic"),
        pytest.param("simple test string", ["simple", "test", "string"], id="no_quotes_simple"),
        pytest.param("singleword", ["singleword"], id="no_quotes_single_word"),
        pytest.param("", [], id="empty_string"),  # Empty string
        pytest.param(
            "   multiple   spaces   ", ["multiple", "spaces"], id="no_quotes_multi_spaces"
        ),  # Leading/trailing/multiple internal spaces
        pytest.param("  ", [], id="only_spaces"),  # Only spaces
        # Cases with single quotes
        pytest.param(
            "I live in 'La guardia'", ["I", "live", "in", "La guardia"], id="single_quotes_basic"
        ),
        pytest.param(
            "'quoted phrase' at start",
            ["quoted phrase", "at", "start"],
            id="single_quotes_at_start",
        ),
        pytest.param(
            "end with 'another one'", ["end", "with", "another one"], id="single_quotes_at_end"
        ),
        pytest.param(
            "middle 'quoted part' here",
            ["middle", "quoted part", "here"],
            id="single_quotes_in_middle",
        ),
        pytest.param("'single'", ["single"], id="single_quotes_only_word"),
        pytest.param(
            "two 'quoted' 'parts'", ["two", "quoted", "parts"], id="single_quotes_multiple"
        ),
        pytest.param("''", [], id="single_quotes_empty"),  # Empty single quoted string
        # Cases with double quotes
        pytest.param(
            'Hello, "World Wide Web" is great!',
            ["Hello,", "World Wide Web", "is", "great!"],
            id="double_quotes_basic",
        ),
        pytest.param(
            '"double quoted" example', ["double quoted", "example"], id="double_quotes_at_start"
        ),
        pytest.param(
            'another "test case" here',
            ["another", "test case", "here"],
            id="double_quotes_in_middle",
        ),
        pytest.param('""', [], id="double_quotes_empty"),  # Empty double quoted string
        # Cases with backticks
        pytest.param(
            "Path is `/usr/local/bin`", ["Path", "is", "/usr/local/bin"], id="backticks_basic"
        ),
        pytest.param(
            "command `ls -la /tmp` to list",
            ["command", "ls -la /tmp", "to", "list"],
            id="backticks_command_example",
        ),
        pytest.param("``", [], id="backticks_empty"),  # Empty backtick string
        # Mixed quotes and complex cases
        pytest.param(
            "Mixed 'quotes and' \"other things\" `like this`",
            ["Mixed", "quotes and", "other things", "like this"],
            id="mixed_quotes",
        ),
        pytest.param(
            "Text with 'single quotes' and \"double quotes\" in it",
            ["Text", "with", "single quotes", "and", "double quotes", "in", "it"],
            id="mixed_types_of_quotes",
        ),
        pytest.param(
            'Path: `/usr/bin/python` Version: "3.9.7"',
            ["Path:", "/usr/bin/python", "Version:", "3.9.7"],
            id="mixed_quotes_with_punctuation",
        ),
        pytest.param(
            "A string with \"empty\" and '' quotes",
            ["A", "string", "with", "empty", "and", "quotes"],
            id="mixed_empty_quotes",
        ),
    ],
)
def test_split_quoted_string(input_string: str, expected_output: list[str]):
    actual_output = split_quoted_string(input_string)
    assert actual_output == expected_output, (
        f"For input: '{input_string}', Expected: {expected_output}, Got: {actual_output}"
    )



import pytest

from ynab_unlinked.config import LATEST_VERSION, MAX_CONFIG_VERSION
from ynab_unlinked.config.core import ConfigError, config_version
from ynab_unlinked.config.migrations.base import Version


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize(
    "config, expected",
    [[f"V{i}", f"V{i}"] for i in range(1, MAX_CONFIG_VERSION + 1)] + [["V0", "V2"]],
    indirect=["config"],
    ids=[f"V{i}" for i in range(1, MAX_CONFIG_VERSION + 1)] + ["invalid_version"],
)
def test_config_verison(expected: str):
    assert config_version() == Version("Config", expected)


# The missing version does not exist in tests assets
@pytest.mark.version("missing")
@pytest.mark.usefixtures("config")
def test_config_version_with_no_file_returns_latest_version():
    assert config_version() == LATEST_VERSION


# The broken version exist but has no version in it
@pytest.mark.version("broken")
@pytest.mark.usefixtures("config")
def test_config_version_without_version_raises():
    with pytest.raises(ConfigError):
        config_version()



import contextlib
import io
from collections.abc import Generator, Sequence
from pathlib import Path
from typing import Any, Literal, overload

from ynab_unlinked import display
from ynab_unlinked.exceptions import ParsingError


@overload
def pdf(
    input_file: Path,
    allow_empty_columns: Literal[False] = False,
    table_settings: dict[str, Any] | None = None,
    expected_number_of_columns: int | None = None,
) -> Generator[Sequence[str]]: ...


@overload
def pdf(
    input_file: Path,
    allow_empty_columns: Literal[True] = True,
    table_settings: dict[str, Any] | None = None,
    expected_number_of_columns: int | None = None,
) -> Generator[Sequence[str | None]]: ...


def pdf(
    input_file: Path,
    allow_empty_columns: bool = False,
    table_settings: dict[str, Any] | None = None,
    expected_number_of_columns: int | None = None,
) -> Generator[Sequence[str | None]]:
    """
    Parse a pdf and extract the main table from it. The table is extracted using
    [pdfplumber](https://github.com/jsvine/pdfplumber).

    Arguments:
    - input_file (Path): the input file to parse
    - table_settings (dict[str, Any]): these are the table settings passed
      to pdfpluber (see [docs](https://github.com/jsvine/pdfplumber/tree/stable?tab=readme-ov-file#table-extraction-settings))

    The method yields elements from a list of rows that contains a list of columns as
    Sequence[str]
    """
    import pdfplumber

    # Capture stderr to process potential non-CropBox messages later
    # This is because the underlying pdf parsing library can generate warnings with
    # no impact that break the usage of the tool
    stderr_capture = io.StringIO()

    with contextlib.redirect_stderr(stderr_capture), pdfplumber.open(input_file) as pdf:
        for page_number, page in enumerate(pdf.pages):
            table = page.extract_table(table_settings=table_settings or {})
            if table is None:
                raise ParsingError(
                    input_file,
                    f"No transaction table was found in page {page_number}",
                )

            for row in table:
                if (
                    expected_number_of_columns is not None
                    and (n_columns := len(row)) != expected_number_of_columns
                ):
                    raise ParsingError(
                        input_file,
                        f"Expected {expected_number_of_columns} but found {n_columns}",
                    )

                if any(c is None for c in row) and not allow_empty_columns:
                    raise ParsingError(
                        input_file,
                        "Malformed Transaction Table: Expected no empty column but found at least one.",
                    )

                yield row

    if captured_output := stderr_capture.getvalue():
        cropbox_message = "CropBox missing from /Page, defaulting to MediaBox"
        if remaining_lines := [
            line for line in captured_output.strip().split("\n") if line.strip() != cropbox_message
        ]:
            remaining_output = "\n".join(remaining_lines)
            display.warning(f"Potential PDF issue reading {input_file}:\n{remaining_output}")



import datetime as dt
from collections.abc import Generator
from pathlib import Path

import typer

from ynab_unlinked import display
from ynab_unlinked.config import ConfigV2
from ynab_unlinked.config.constants import TRANSACTION_GRACE_PERIOD_DAYS
from ynab_unlinked.config.models.shared import Checkpoint, EntityConfig
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.display import bullet_list, confirm, console, info, process, question
from ynab_unlinked.entities import Entity
from ynab_unlinked.exceptions import ParsingError
from ynab_unlinked.matcher import match_transactions
from ynab_unlinked.models import MatchStatus, Transaction, TransactionWithYnabData
from ynab_unlinked.payee import set_payee_from_ynab
from ynab_unlinked.utils import (
    display_partial_matches,
    display_transaction_table,
    display_transactions_to_upload,
)
from ynab_unlinked.ynab_api.client import Client

# Request transactions to the YNAB API from the last checkpoint date minus 10 days for buffer
TRANSACTIONS_DAYES_BEFORE_LAST_EXTRACTION = 10


def add_past_to_transactions(transactions: list[Transaction], checkpoint: Checkpoint | None):
    if checkpoint is None:
        return

    for t in transactions:
        if t.date < checkpoint.latest_date_processed + dt.timedelta(
            days=TRANSACTION_GRACE_PERIOD_DAYS
        ):
            t.past = True


def add_counter_to_existing_transactions(transactions: list[Transaction]):
    """
    This method check every transaction that has the same date, payee and amount
    and increments its counter to ensure that they have a unique import ID when
    being added to YNAB.
    """
    counters: dict[str, int] = {}
    for t in transactions:
        if t.id not in counters:
            counters[t.id] = 0
        else:
            counters[t.id] += 1
            t.counter = counters[t.id]


def preprocess_transactions(transactions: list[Transaction], checkpoint: Checkpoint | None):
    add_past_to_transactions(transactions, checkpoint)
    add_counter_to_existing_transactions(transactions)


def filter_transactions(
    transactions: list[Transaction], checkpoint: Checkpoint | None
) -> Generator[Transaction]:
    # Checkpoint filtering has many issues when dealing with old trasactions that took time to
    # be processed. While we rethink it lets just not filter them
    # if checkpoint is None:
    yield from transactions
    return

    # yield from (t for t in transactions if t.date >= checkpoint.latest_date_processed)


def get_or_prompt_account_id(config: ConfigV2, entity_name: str, force_prompt: bool) -> str:
    if entity_name in config.entities and not force_prompt:
        return config.entities[entity_name].account_id

    display.info(f"Lets select the account for {entity_name.capitalize()}:")
    client = Client(config.api_key)
    budget_id = config.budget.id

    accounts = [acc for acc in client.accounts(budget_id=budget_id) if not acc.closed]

    console().print(bullet_list(f"{idx + 1:>2}: {acc.name}" for idx, acc in enumerate(accounts)))

    acc_num = question(
        "What account are the transactions going to be imported to? (By number)",
        choices=[str(i) for i in range(1, len(accounts) + 1)],
        show_choices=False,
    )
    account = accounts[int(acc_num) - 1]

    info(f"Account selected: {account.name}")

    if not force_prompt:
        config.entities[entity_name] = EntityConfig(account_id=account.id)
        config.save()

    return account.id


def process_transactions(
    entity: Entity,
    input_file: Path,
    context: YnabUnlinkedContext,
) -> None:
    """
    Process the transactions from the input file, match them with YNAB data, and upload them to YNAB.

    If the Entity calling this method does not have a config stored, the user will be prompted to select an account
    this entity should publish transactions to. This account will be used moving forward when using this entity.

    If the user called `yul -a` the user will always be promptped to select
    and account and the selected account won't be saved for this particular entity.
    """

    config = context.config
    show = context.show
    reconcile = context.reconcile

    acount_id = get_or_prompt_account_id(config, entity.name(), force_prompt=context.choose_account)

    try:
        parsed_input = entity.parse(input_file, context)
    except ParsingError as e:
        display.error(f"Error when parsing {e.input_file}")
        display.console().print(f"  Message: {e.message}")
        raise typer.Exit(1) from e

    checkpoint = config.entities[entity.name()].checkpoint

    preprocess_transactions(parsed_input, checkpoint)

    if show:
        display_transaction_table(parsed_input, context.formatter)
        return

    transactions = [
        TransactionWithYnabData(t)
        for t in filter_transactions(
            parsed_input,
            checkpoint,
        )
    ]

    client = Client(config.api_key)
    budget_id = config.budget.id

    earliest_transaction = min(t.date for t in transactions)

    with process("Reading transactions..."):
        ynab_transactions = client.transactions(
            budget_id=budget_id,
            account_id=acount_id,
            since_date=earliest_transaction - dt.timedelta(days=context.buffer),
        )
    display.success("✔ Transactions read")

    with process("Augmenting transactions..."):
        match_transactions(transactions, ynab_transactions, reconcile, config)
        set_payee_from_ynab(transactions, client, config)
    display.success("✔ Transactions augmneted with YNAB information")

    display_transactions_to_upload(transactions, context.formatter)

    if not any(t.needs_creation for t in transactions):
        info("🎉 All done! Nothing to do.")
        if transactions:
            config.update_and_save(transactions[0], entity.name())
        return

    if partial_matches := [
        t for t in transactions if t.match_status is MatchStatus.PARTIAL_MATCH and t.needs_creation
    ]:
        display_partial_matches(partial_matches, context.formatter)
        display.info(
            "\nIf these partial matches are ok, you can accept them and we will keep track of the "
            "payee name for future reference."
        )
        final_matching = (
            MatchStatus.MATCHED
            if confirm("Do you want to accept these matches?")
            else MatchStatus.UNMATCHED
        )
        for t in partial_matches:
            t.match_status = final_matching
            if final_matching == MatchStatus.UNMATCHED:
                t.reset_matching()

        if final_matching == MatchStatus.MATCHED:
            # If the user agreed to match these, add renaming rules to config
            config.add_payee_rules(partial_matches)

    with process("Preparing transactions to upload..."):
        new_transactions = [t for t in transactions if t.needs_creation]

    display.info(f"Transactions to import:       {len(new_transactions)}")

    if confirm("Do you want to continue and create the transactions?"):
        with process("Creating/Updating transactions..."):
            client.create_transactions(
                budget_id=budget_id,
                account_id=acount_id,
                transactions=new_transactions,
            )

        config.update_and_save(transactions[0], entity.name())

    display.info("🎉 All done!")



from pathlib import Path
from typing import Protocol

from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.models import Transaction


class Entity(Protocol):
    def parse(self, input_file: Path, context: YnabUnlinkedContext) -> list[Transaction]:
        """
        Parse an input file into a list of Transaction objects.

        This is the main method of the EntityParser protocol. Any input file can be converted
        into an abstraction of Transaction objects. These objects only contain information
        related with the transactions themselves:
        - Date
        - Payee
        - Amount

        `ynab-unlinked` will understand these transactions and enrich them when necesary to
        ensure the best matching when pushing them to YNAB.
        """
        ...

    def name(self) -> str:
        """
        Returns the name of the entity.
        """
        ...



from typing import Final, cast

import typer

from ynab_unlinked import app
from ynab_unlinked.commands import config_app, load
from ynab_unlinked.config import Config, ConfigV1, ConfigV2, get_config
from ynab_unlinked.config.core import ConfigError
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.display import bold, success
from ynab_unlinked.formatter import Formatter
from ynab_unlinked.utils import prompt_for_api_key, prompt_for_budget

app.add_typer(load, name="load")
app.add_typer(config_app, name="config")

VERSION_MAPPING: Final[dict[str, type[Config]]] = {
    "V1": ConfigV1,
    "V2": ConfigV2,
}


@app.command(name="setup")
def setup_command():
    """Setup YNAB Unlinked"""
    bold("Welcome to ynab-unlinked! Lets setup your connection")
    api_key = prompt_for_api_key()
    budget = prompt_for_budget(api_key)
    config = ConfigV2(api_key=api_key, budget=budget)
    config.save()

    success("All done!")


@app.callback(no_args_is_help=True)
def cli(context: typer.Context):
    """
    Create transations in your YNAB account from a bank export of your extract.
    \n

    The first time the command is run you will be asked some questions to setup your YNAB connection. After that,
    transaction processing won't require any input unless there are some actions to take for specific transactions.
    """

    if context.invoked_subcommand == "setup":
        # If we are running setup there is nothing to do here
        return

    # Load the proper config or run setup to create it
    config = get_config()
    if config is None:
        setup_command()
        config = get_config()

    if config is None:
        raise ConfigError("Unexpected error: config could not be loaded")

    context.obj = YnabUnlinkedContext(
        config=cast(ConfigV2, config),
        extras=None,
        formatter=Formatter(
            date_format=config.budget.date_format,
            currency_format=config.budget.currency_format,
        ),
    )


def main():
    app(prog_name="yul")



from __future__ import annotations

import datetime as dt
from pathlib import Path

from pydantic import BaseModel, Field

from ynab_unlinked.config.constants import TRANSACTION_GRACE_PERIOD_DAYS
from ynab_unlinked.config.migrations import Version
from ynab_unlinked.config.paths import config_path
from ynab_unlinked.models import Transaction, TransactionWithYnabData

from .shared import Checkpoint, EntityConfig


class ConfigV1(BaseModel):
    api_key: str
    budget_id: str
    last_reconciliation_date: dt.date | None = None
    entities: dict[str, EntityConfig] = Field(default_factory=dict)
    payee_rules: dict[str, set[str]] = Field(default_factory=dict)

    @staticmethod
    def version() -> Version:
        return Version("Config", "V1")

    @staticmethod
    def path() -> Path:
        return config_path(ConfigV1.version().version)

    def save(self):
        self.path().parent.mkdir(parents=True, exist_ok=True)
        self.path().write_text(self.model_dump_json(indent=4))

    def update_and_save(self, last_transaction: Transaction, entity_name: str):
        checkpoint = Checkpoint(
            latest_date_processed=(
                last_transaction.date - dt.timedelta(days=TRANSACTION_GRACE_PERIOD_DAYS)
            ),
            latest_transaction_hash=hash(last_transaction),
        )

        self.entities[entity_name].checkpoint = checkpoint

        self.save()

    @staticmethod
    def load() -> ConfigV1:
        return ConfigV1.model_validate_json(ConfigV1.path().read_text())

    @staticmethod
    def exists() -> bool:
        return ConfigV1.path().is_file()

    def add_payee_rules(self, transactions: list[TransactionWithYnabData]):
        # For each transaction, add a rule that matches both payees
        for transaction in transactions:
            if transaction.partial_match is None:
                continue

            if transaction.ynab_payee is None:
                continue

            imported_payee = transaction.payee
            ynab_payee = transaction.ynab_payee

            if imported_payee == ynab_payee:
                continue

            self.payee_rules.setdefault(ynab_payee, set()).add(imported_payee)
            self.save()

    def payee_from_fules(self, payee: str) -> str | None:
        return next(
            (
                ynab_payee
                for ynab_payee, valid_names in self.payee_rules.items()
                if payee in valid_names
            ),
            None,
        )

    def entity(self, name: str) -> EntityConfig | None:
        return self.entities.get(name)

    def set_entity_account(self, name: str, account_id: str):
        if (entity := self.entities.get(name)) is not None:
            entity.account_id = account_id



from __future__ import annotations

from pathlib import Path
from typing import Annotated

import typer


def command(
    context: typer.Context,
    input_file: Annotated[
        Path,
        typer.Argument(exists=True, file_okay=True, dir_okay=False, readable=True),
    ],
):
    """
    Inputs transactions from a Sabadell TXT or XLS file.

    From your Sabadell Credit Card statement, you can download a txt file with the transactions.
    At the moment only txt format is supported.
    """

    from ynab_unlinked.context_object import YnabUnlinkedContext
    from ynab_unlinked.process import process_transactions

    from .sabadell import SabadellParser

    ctx: YnabUnlinkedContext = context.obj

    process_transactions(
        entity=SabadellParser(),
        input_file=input_file,
        context=ctx,
    )



from dataclasses import dataclass

from ynab_unlinked.config import ConfigV2
from ynab_unlinked.formatter import Formatter


@dataclass
class YnabUnlinkedContext[T]:
    config: ConfigV2
    formatter: Formatter
    extras: T
    show: bool = False
    reconcile: bool = False
    choose_account: bool = False
    buffer: int = 15



from collections.abc import Generator, Sequence
from pathlib import Path
from typing import Any

ASSETS_PATH = Path(__file__).parents[1] / "assets"
AssetPath = Path | str


def path(asset_path: AssetPath) -> Path:
    return ASSETS_PATH / asset_path


def read_text(asset_path: AssetPath, encoding="utf-8") -> str:
    return path(asset_path).read_text(encoding=encoding)


def read_lines(asset_path: AssetPath, encoding="utf-8") -> Generator[str]:
    with open(path(asset_path), encoding=encoding) as asset_file:
        yield from asset_file


def read_pdf(
    asset_path: AssetPath,
    allow_empty_columns: bool = False,
    table_settings: dict[str, Any] | None = None,
    expected_number_of_columns: int | None = None,
) -> Generator[Sequence[str]]:
    from ynab_unlinked.parsers import pdf

    yield from pdf(
        path(asset_path),
        allow_empty_columns=allow_empty_columns,  # type: ignore since we do not want to overload this method
        table_settings=table_settings,
        expected_number_of_columns=expected_number_of_columns,
    )


def read_xls(
    asset_path: AssetPath,
    read_after_row: int = 0,
    read_after_row_like: Sequence[str] | None = None,
) -> Generator[Sequence[str]]:
    from ynab_unlinked.parsers import xls

    yield from xls(
        path(asset_path),
        read_after_row=read_after_row,
        read_after_row_like=read_after_row_like,
    )



from pathlib import Path


class ParsingError(Exception):
    def __init__(self, input_file: Path, message: str):
        self.input_file = input_file
        self.message = message
        super().__init__(message)



from __future__ import annotations

from collections.abc import Iterable
from contextlib import contextmanager

from rich.console import Console
from rich.prompt import Confirm, Prompt
from rich.status import Status

__console: Console | None = None


def console() -> Console:
    global __console
    if __console is None:
        __console = Console()
    return __console


def success(message: str):
    console().print(f"[bold green]{message}[/bold green]")


def error(message: str):
    console().print(f"[bold red]{message}[/bold red]")


def warning(message: str):
    console().print(f"[bold orange1]{message}[/bold orange1]")


def info(message: str):
    console().print(message)


def bold(message: str):
    return console().print(f"[bold]{message}[/bold]")


@contextmanager
def process(message: str, completed_message: str | None = None):
    with Status(message, spinner="dots"):
        yield
    if completed_message:
        success(completed_message)


def question(message: str, **kwargs) -> str:
    return Prompt.ask(f"[bold cyan]{message}[/bold cyan]", console=console(), **kwargs)


def confirm(message: str, **kwargs) -> bool:
    return Confirm.ask(f"[bold cyan]{message}[/bold cyan]", console=console(), **kwargs)


def bullet_list(items: Iterable[str]) -> str:
    return "\n".join(f" • {item}" for item in items)



import datetime as dt

import pytest

from tests.helpers.types import CliRunner, LoadEntityCallback
from tests.helpers.ynab_api import YnabClientStub

pytestmark = pytest.mark.version("V2")


def test_load(
    yul: CliRunner,
    load_entity: LoadEntityCallback,
    today: dt.datetime,
    ynab_api: YnabClientStub,
):
    load_entity(today)
    result = yul("load --show test")
    assert result.exit_code == 0, f"Error found: {result.output_bytes}"
